package vocleditor.command;

import org.eclipse.gef.commands.Command;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Shell;

import vocleditor.dialogs.CreateAssociationRole;
import vocleditor.model.VOCLModelManager;
import VOCLModel.AssociationEndRoleSource;
import VOCLModel.AssociationEndRoleTarget;
import VOCLModel.AssociationRole;
import VOCLModel.ClassifierNode;
import VOCLModel.NavExp;
import VOCLModel.VOCLModelFactory;

/**
 * 
 * Class       : AssociationRoleConnectionCommand<br/>
 * Package     : vocleditor.command<br/>
 * Project     : VOCLEditor<br/>
 * Created On  : 22.05.2004<br/>
 * Description : Command for creating an association between two ClassifierNodes.<br/>
 * @author Gruppe 1: Stefan Hnsgen, 
 * 					 Andre Hernich,
 * 			 		 Maximilian Kern,
 *		 			 Sebastian Russin<br>
 */
public class AssociationRoleConnectionCommand extends Command {

	// Description shown in the editor
	private static final String description = "Connects Classifiers";
	// Label to show in editor palette
	private static final String label = "AssociationRole";
	// the new association
	protected AssociationRole link;

	//parent of the association (always a navexpression)
	protected NavExp parent;

	// Variables to remember source, target, and link. This is neccessary for 
	// execute().
	protected ClassifierNode source;
	protected ClassifierNode target;
	protected AssociationEndRoleSource s;
	protected AssociationEndRoleTarget t;
	
	protected Shell shell;

	/**
	 * Constructor
	 */
	public AssociationRoleConnectionCommand() {
		super(label);
	}

	/**
	 * Flag that indicates whether the command can be executed at the moment of
	 * the call.
	 * 
	 * @return true, if it can be executed, otherwise false.
	 */
	public boolean canExecute() {
		if ((source == null || target == null || link == null))
			return false;
		else if (source == target) {
			return false;
		} else if (source.getNavExp() != target.getNavExp())
			return false;
		else
			return true;
	}

	/**
	 * Executes the command.
	 */
	public void execute() {
		CreateAssociationRole d = new CreateAssociationRole(shell,SWT.NULL);
		d.setSource(source);
		d.setTarget(target);
		d.open();
		if(d.isCancelled())
			return;
		
		NavExp navExp = source.getNavExp();
		//set NavExp as Parent
		setParent(navExp);
		//set NavExp and a Name for the AssociationRole
		link.setNavExp(navExp);
		link.setName(source.getName() + "-->" + target.getName());
		//add AssociationRole into Modell
		parent.getAssociationRole().add(link);

		VOCLModelFactory mfc = VOCLModelManager.getFactory();
		//Create AssociationEndRoles
		s = mfc.createAssociationEndRoleSource();
		t = mfc.createAssociationEndRoleTarget();
		//add AssociationEndRoles to the AssociationRole
		link.setAssociationEndRoleSource(s);
		link.setAssociationEndRoleTarget(t);
		
		//set the names for the Labels displayed at the ends of an association
		t.setName(d.getTargetName().toLowerCase());
		s.setName("");

		//insert AssociationEndRoles into Modell
		s.setClassifierNode(source);
		t.setClassifierNode(target);
	}

	/**
	 * Returns the description for this command.
	 */
	public String getLabel() {
		return description;
	}

	/**
	 * Returns the Associationrole itself.
	 */
	public AssociationRole getLink() {
		return link;
	}
	/**
	 * Method for returning the parent of the AssociationRole.
	 * @return parent of the Associationrole
	 */
	public NavExp getParent() {
		return parent;
	}

	/**
	 * Returns the source classifiernode of the associationrole.
	 */
	public ClassifierNode getSource() {
		return source;
	}

	/**
	 * Returns the target classifiernode of the associationrole.
	 */
	public ClassifierNode getTarget() {
		return target;
	}

	/**
	 * Executes the command after undo.
	 */
	public void redo() {
		source.getAssociationEndRole().add(s);
		target.getAssociationEndRole().add(t);
		getParent().getAssociationRole().add(link);
	}

	/**
	 * Sets the associationrole.
	 */
	public void setAssociationRole(AssociationRole newLink) {
		link = newLink;
	}

	/**
	 * Set the current parent.
	 * @param exp
	 */
	public void setParent(NavExp exp) {
		parent = exp;
	}

	/**
	 * Sets the source classifiernode for this associationrole.
	 */
	public void setSource(ClassifierNode newSource) {
		source = newSource;
	}

	/**
	 * Sets the target classifiernode for this associationrole.
	 */
	public void setTarget(ClassifierNode newTarget) {
		target = newTarget;
	}

	/**
	 * Restores the state before execution of the command.
	 */
	public void undo() {
		source.getAssociationEndRole().remove(s);
		target.getAssociationEndRole().remove(t);
		getParent().getAssociationRole().remove(link);
	}

	/**
	 * @return Returns the shell.
	 */
	protected Shell getShell() {
		return shell;
	}
	/**
	 * @param shell The shell to set.
	 */
	public void setShell(Shell shell) {
		this.shell = shell;
	}
}
