package vocleditor.command;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.gef.commands.Command;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;

import vocleditor.dialogs.NewConditionDialog;
import vocleditor.model.Helpers;
import vocleditor.model.VOCLModelManager;
import VOCLModel.Attribute;
import VOCLModel.ClassifierNode;
import VOCLModel.ClassifierRole;
import VOCLModel.ClassifierRoleSet;
import VOCLModel.Condition;
import VOCLModel.Constraint;
import VOCLModel.Method;
import VOCLModel.NavExp;
import VOCLModel.OpKindEnum;
import VOCLModel.VOCLExpression;
import VOCLModel.VOCLModelFactory;
import VOCLModel.VariableDeclaration;

/**
 * 
 * Class       : CreateConditionCommand<br/>
 * Package     : vocleditor.command<br/>
 * Project     : VOCLEditor<br/>
 * Created On  : 19.06.2004<br/>
 * Description : Command for Creating a new Condition.<br/>
 * @author Gruppe 1: Stefan Hnsgen, 
 * 					 Andre Hernich,
 * 			 		 Maximilian Kern,
 *		 			 Sebastian Russin<br>
 */
public class CreateConditionCommand extends Command {

	//for displaying a dialog
	Shell shell;
	//the current condition
	Condition condition;
	//parent of the condition (Navexp or Constraint)
	Object parent;
	//variable on the left side
	VariableDeclaration left = null;
	//variable on the right side
	VariableDeclaration right = null;

	//the parent of the navexp that is the parent of this condition
	//needed for drawing conditions without classifiers in the current
	//navexp
	private Object navexpParent;
	//the indexed-position of the navexpfigure
	private int position;

	private static final String label = "Condition";

	private static final String description = "Creates a Condition";

	/**
	 * Constructor for the condition
	 * @param shell the shell for displaying dialogs
	 */
	public CreateConditionCommand(Shell shell) {
		super(label);
		this.shell = shell;
	}

	/**
	 * Set the current condition
	 * @param condition
	 */
	public void setCondition(Condition condition) {
		this.condition = condition;
	}

	/**
	 * Set the parent of the condition.
	 * @param parent parent that is either a navexp or a constraint.
	 */
	public void setParent(Object parent) {
		this.parent = parent;
	}

	/**
	 * @see org.eclipse.gef.commands.Command#execute()
	 */
	public void execute() {

		if (parent instanceof Constraint) {
			//if the user wants to create a Condition for a constraint, although 
			//there exists a BoolExp, If-then-else or let-Expression
			 if (!(((Constraint) parent).getVOCLExpression()
				instanceof NavExp)) {
				MessageBox box = new MessageBox(shell, SWT.ICON_WARNING);
				box.setText("Fehler beim Erstellen");
				box.setMessage(
					"Es knnen Conditions nur innerhalb einer NavExp erstellt werden");
				box.open();
				return;
			}

			//check, if there exists already one Condition
			if (((Constraint) parent).getCondition() != null) {
				MessageBox box = new MessageBox(shell, SWT.ICON_WARNING);
				box.setText("Fehler beim erstellen einer Condition!!!");
				box.setMessage(
					"Es existiert bereits eine Condition!\n"
						+ "Lschen Sie diese zuerst");
				box.open();
				return;
			}

			//open a conditon dialog
			NewConditionDialog d = new NewConditionDialog(shell, 0);
			//set variablen to the dialog
			List variablen =
				collectVariablesOfNavexp(
					(NavExp) ((Constraint) parent).getVOCLExpression());
			d.setLinkeVariablen(variablen);
			d.open();
			if (d.isCanceled()) {
				return;
			}

			//create a new Condition
			VOCLModelFactory mfc = VOCLModelManager.getFactory();

			//search the right VariableDeclaration selected on the left side of the dialog
			for (Iterator it = variablen.iterator(); it.hasNext();) {
				VariableDeclaration var = (VariableDeclaration) it.next();
				if (var.getVarName().equals(d.getLeftName())) {
					left = var;
				}
			}

			//set a reference to the condition
			if (left != null) {
				condition.setLeft(left);
			}
			//set the operation of the condition
			condition.setOp(d.getOp());

			//if a Value was select for the right side
			if (d.getRightType() == null) {
				condition.setValue(d.getRightName());
			} else {
				//search a reference on the instance of the variabledeclaration
				for (Iterator it = variablen.iterator(); it.hasNext();) {
					VariableDeclaration var = (VariableDeclaration) it.next();
					if (var.getVarName().equals(d.getRightName())) {
						right = var;
					}
				}
				condition.setRight(right);
				condition.setValue("");
			}
			//insert condition into modell
			((Constraint) parent).setCondition(condition);

		} else {
			//parent is a Navexp
			//open a conditon dialog
			VOCLModelFactory mfc = VOCLModelManager.getFactory();
			if (parent == null) {
				parent = mfc.createNavExp();
				((NavExp) parent).setIndex(position);
				Helpers.addToParent(
					(VOCLExpression) parent,
					navexpParent,
					position);
			}
			NewConditionDialog d = new NewConditionDialog(shell, 0);
			//collect all variables for displaying in the dialog
			List variablen = collectVariablesOfNavexp((NavExp) parent);
			d.setLinkeVariablen(variablen);
			//open the dialog
			d.open();
			if (d.isCanceled()) {
				Helpers.removeFromParent(navexpParent, position);
				return;
			}

			//search the right VariableDeclaration selected on the left side of the dialog
			for (Iterator it = variablen.iterator(); it.hasNext();) {
				VariableDeclaration var = (VariableDeclaration) it.next();
				if (var.getVarName().equals(d.getLeftName())) {
					left = var;
				}
			}

			//set a reference to the condition
			if (left != null) {
				condition.setLeft(left);
			}
			//set the operation of the condition
			condition.setOp(d.getOp());

			//if a Value was select for the right side
			if (d.getRightType() == null) {
				condition.setValue(d.getRightName());
			} else {
				//set a reference on the instance of the variabledeclaration
				for (Iterator it = variablen.iterator(); it.hasNext();) {
					VariableDeclaration var = (VariableDeclaration) it.next();
					if (var.getVarName().equals(d.getRightName())) {
						right = var;
					}
				}
				condition.setRight(right);
				condition.setValue("");
			}
			//insert condition into modell
			((NavExp) parent).setCondition(condition);
		}

	}

//	private VariableDeclaration getLetExpVarDec(NavExp nav) {
//		EObject object = nav.eContainer();
//		while (true) {
//			if (object instanceof LetExp) {
//				return ((LetExp) object).getVariableDeclaration();
//			} else if (object instanceof Constraint) {
//				return null;
//			}
//			object = object.eContainer();
//		}
//	}
	
	
	/**
	 * Returns a list of VariableDeclarations of attributevariables, return-variable of methods 
	 * and VariableDeclarations of a LetExp if it exists.
	 * @param nav the parent of this condition
	 * @return the list of VariableDeclarations
	 */
	private List collectVariablesOfNavexp(NavExp nav) {
		List variablen = new LinkedList();
		//if there exists a LetExp as parent of this parent(navexp), add the VariableDeclaration
		if (Helpers.getLetExpVariablenDeclaration(nav) != null) {
			variablen.add(Helpers.getLetExpVariablenDeclaration(nav));
		}
		//select all ClassifierNodes	
		for (Iterator it = nav.getClassifierNode().iterator(); it.hasNext();) {
			ClassifierNode classifier = (ClassifierNode) it.next();
			//select all Variabledeclaration of the attributes
			for (Iterator attrIt = classifier.getAttribute().iterator();
				attrIt.hasNext();
				) {
				Attribute temp = (Attribute) attrIt.next();
				if (temp.getVariableDeclaration() != null && temp.getVariableDeclaration().getOpKindEnumIterator().isEmpty()) {
					variablen.add(temp.getVariableDeclaration());
					
				}
			}
			if (classifier instanceof ClassifierRole) {
				//add the return type of a method to the variablenlist
				if (((ClassifierRole) classifier).getMethod() != null) {
					Method m = ((ClassifierRole) classifier).getMethod();
					variablen.add(m.getReturns());
				}
			} else if (classifier instanceof ClassifierRoleSet) {
				//add the call variable of a sum / size operation to the variable list
				OpKindEnum kindEnum =
					((ClassifierRoleSet) classifier).getOpKindEnum();
				if (kindEnum != null) {
					VariableDeclaration callVar = kindEnum.getCallVar();
					if (callVar != null) {
						variablen.add(callVar);
					}
				}
			}

		}
		return variablen;
	}

	/**
	 * @see org.eclipse.gef.commands.Command#undo()
	 */
	public void undo() {
		//set the reference to the VariableDeclarations to null
		condition.setLeft(null);
		condition.setRight(null);
		//remove condition from modell
		if (parent instanceof NavExp)
			 ((NavExp) parent).setCondition(null);
		else
			 ((Constraint) parent).setCondition(null);
	}

	/**
	 * @see org.eclipse.gef.commands.Command#redo()
	 */
	public void redo() {
		condition.setLeft(left);
		condition.setRight(right);
		if (parent instanceof NavExp) {
			((NavExp) parent).setCondition(condition);
		} else {
			((Constraint) parent).setCondition(condition);
		}
	}
	/**
	 * Set the index - position of the parent (navexp) 
	 * @param i the index.
	 */
	public void setPosition(int i) {
		position = i;
	}

	/**
	 * Sets the current parent if it already exists.
	 * @param object
	 */
	public void setNavexpParent(Object object) {
		navexpParent = object;
	}

}
