package vocleditor.command;

import java.util.List;

import org.eclipse.draw2d.geometry.Point;
import org.eclipse.gef.commands.Command;
import org.eclipse.swt.widgets.Shell;

import vocleditor.dialogs.NewConstraintDialog;
import vocleditor.model.VOCLModelManager;
import VOCLModel.Constraint;
import VOCLModel.ConstraintKind;
import VOCLModel.Context;
import VOCLModel.VOCLDiagram;
import VOCLModel.VOCLModelFactory;
import VOCLModel.VariableDeclaration;

/**
 * 
 * Class       : CreateNewConstraintCommand<br/>
 * Package     : vocleditor.command<br/>
 * Project     : VOCLEditor<br/>
 * Created On  : 24.06.2004<br/>
 * Description : Command for Creating a new Constraint.<br/>
 * @author Gruppe 1: Stefan Hnsgen, 
 * 					 Andre Hernich,
 * 			 		 Maximilian Kern,
 *		 			 Sebastian Russin<br>
 */
public class CreateNewConstraintCommand extends Command {

	// Label to show in editor palette
	private static final String label = "Constraint";

	// Description shown in the editor
	private static final String description = "Creates a new Constraint";

	// Data for execute()
	private Constraint constraint = null;
	private VariableDeclaration variableDeclaration = null;
	private List attributes = null;
	private Point loc;
	private VOCLDiagram parent;

	// To pop up the dialog, we need a reference to the parent shell...
	private Shell shell;

	/**
	 * Creates the command.
	 */
	public CreateNewConstraintCommand() {
		super(label);
	}

	/**
	 * Creates the command and uses the given parent shell to display the 
	 * dialog box.
	 * 
	 * @param shell the shell.
	 */
	public CreateNewConstraintCommand(Shell shell) {
		super(label);
		this.shell = shell;
	}

	/**
	 * Sets the parent of the Constraint.
	 * @param newParent the parent (VOCLDiagram)
	 */
	public void setParent(VOCLDiagram newParent) {
		parent = newParent;
	}

	/**
	 * Sets the current location of the Constraint.
	 * @param loc the location
	 */
	public void setLocation(Point loc) {
		this.loc = loc;
	}

	/**
	 * @see org.eclipse.gef.commands.Command#execute()
	 */
	/**
	 * @see org.eclipse.gef.commands.Command#execute()
	 */
	public void execute() {
		NewConstraintDialog d = new NewConstraintDialog(shell);
		d.open();
		if (d.isCancelled())
			return;

		String name = d.getName();
		String contextName = d.getContext_name();
		String contextTyp = d.getContext_type();
		String type = d.getType();

		VOCLModelFactory mfc = VOCLModelManager.getFactory();

		//Context erstellen 
		Context newContext = mfc.createContext();
		newContext.setName(contextName);
		newContext.setType(contextTyp);
		newContext.setIsClassifier(true);
		constraint.setConstraintContext(newContext);

		//als VariablenDeclaration erstellen
		if (!contextName.equals("")) {
			VariableDeclaration contextVar = mfc.createVariableDeclaration();
			contextVar.setVarName(contextName);
			contextVar.setType(contextTyp);
			newContext.getVariableDeclaration().add(contextVar);
			constraint.getConstraintContext().getVariableDeclaration().add(
				contextVar);
			for (int i = 0; i < contextName.length() - 1; i++) {
				if (contextName.substring(i, i + 1).equals(":")) {
					newContext.setIsClassifier(false);
					break;
				}
			}
		}

		constraint.setConstraintContext(newContext);
		constraint.setName(name);
		type = type.trim().toLowerCase();
		if (type.equals("pre"))
			constraint.setKind(ConstraintKind.PRE_LITERAL);
		else if (type.equals("post"))
			constraint.setKind(ConstraintKind.POST_LITERAL);
		else
			constraint.setKind(ConstraintKind.INV_LITERAL);

		//insert all into modell
		insertIntoModel();

		//set the current location of the constraint
		if (loc != null) {
			constraint.setX(loc.x);
			constraint.setY(loc.y);
		}
	}

	/**
	 * Insert all elements into modell.
	 */
	private void insertIntoModel() {
		parent.getConstraint().add(constraint);
		//add variable of context into modell
		if (variableDeclaration != null)
			constraint.getConstraintContext().getVariableDeclaration().add(
				variableDeclaration);
	}

	/**
	 * @see org.eclipse.gef.commands.Command#redo()
	 */
	public void redo() {
		insertIntoModel();
	}

	/**
	 * Sets the current Constraint.
	 * @param cons the current Constraint
	 */
	public void setConstraint(Constraint cons) {
		constraint = cons;
	}

	/**
	 * @see org.eclipse.gef.commands.Command#undo()
	 */
	public void undo() {
		parent.getConstraint().remove(constraint);
	}
}
