
package vocleditor.edit.parts;

import java.util.List;
import java.util.Vector;

import org.eclipse.draw2d.ConnectionAnchor;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.Notifier;
import org.eclipse.gef.ConnectionEditPart;
import org.eclipse.gef.EditPolicy;
import org.eclipse.gef.GraphicalEditPart;
import org.eclipse.gef.NodeEditPart;
import org.eclipse.gef.Request;
import org.eclipse.gef.editparts.AbstractGraphicalEditPart;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.views.properties.IPropertySource;

import vocleditor.edit.policies.ConstraintEditPolicy;
import vocleditor.edit.policies.ConstraintLayoutEditPolicy;
import vocleditor.figures.ConstraintFigure;
import vocleditor.model.EObjectPropertySource;
import VOCLModel.Condition;
import VOCLModel.Constraint;
import VOCLModel.Context;
import VOCLModel.VOCLExpression;
import VOCLModel.VOCLModelPackage;

/**
 * Class       : ConstraintEditPart<br/>
 * Package     : vocleditor.edit.parts<br/>
 * Project     : VOCLEditor<br/>
 * Created On  : 22.05.2004<br/>
 * Description : This editpart represents a constraint<br/>
 * <br/>
 * @author Gruppe 1: Stefan Hnsgen, 
 * 					 Andre Hernich,
 * 			 		 Maximilian Kern,
 *		 			 Sebastian Russin<br>
 *
 *
 */
public class ConstraintEditPart
	extends AbstractGraphicalEditPart
	implements NodeEditPart, Adapter {

	private Notifier target;
	private IPropertySource propertySource = null;
	private Shell shell;

	/**
	 * Creates a new ConstraintEditPart instance.
	 * 
	 * @param classifierRole The classifier role model.
	 */
	public ConstraintEditPart(Constraint classifierRole) {
		setModel(classifierRole);
	}

	/**
	 * Creates the figure.
	 * 
	 * @return The Constraintfigure.
	 */
	protected IFigure createFigure() {
		ConstraintFigure figure = new ConstraintFigure();
		return figure;
	}

	/**
	 * Returns the classifierRole figure.
	 */
	public ConstraintFigure getConstraintFigure() {
		return (ConstraintFigure) getFigure();
	}

	/**
	 * Creates edit policies.
	 */
	protected void createEditPolicies() {

		installEditPolicy(
			EditPolicy.LAYOUT_ROLE,
			new ConstraintLayoutEditPolicy());
		installEditPolicy(
			EditPolicy.COMPONENT_ROLE,
			new ConstraintEditPolicy());
	}

	/**
	 * @see org.eclipse.gef.NodeEditPart#getTargetConnectionAnchor(org.eclipse.gef.Request)
	 */
	public ConnectionAnchor getTargetConnectionAnchor(Request request) {
		return null;
	}

	/**
	 * @see org.eclipse.emf.common.notify.Adapter#getTarget()
	 */
	public Notifier getTarget() {
		return target;
	}

	/**
	 * @see org.eclipse.emf.common.notify.Adapter#setTarget(org.eclipse.emf.common.notify.Notifier)
	 */
	public void setTarget(Notifier newTarget) {
		target = newTarget;
	}

	/**
	 * Return the constraint represented by this edit part
	 */
	public Constraint getConstraint() {
		return (Constraint) getModel();
	}
	/**
	 * @see org.eclipse.gef.editparts.AbstractGraphicalEditPart#activate()
	 */
	public void activate() {
		if (isActive())
			return;
		((Notifier) getConstraint()).eAdapters().add(this);
		VOCLExpression expr = getConstraint().getVOCLExpression();
		if (expr != null)
			expr.eAdapters().add(this);
		Condition cond = getConstraint().getCondition();
		if (cond != null) {
			cond.eAdapters().add(this);
		}
		super.activate();
	}

	/**
	 * @see org.eclipse.gef.editparts.AbstractGraphicalEditPart#deactivate()
	 */
	public void deactivate() {
		if (!isActive())
			return;
		((Notifier) getConstraint()).eAdapters().remove(this);
		VOCLExpression expr = getConstraint().getVOCLExpression();
		if (expr != null)
			expr.eAdapters().remove(this);
		Condition cond = getConstraint().getCondition();
		if (cond != null) {
			cond.eAdapters().remove(this);
		}
		super.deactivate();
	}

	/**
	 * Catch notifications for changes in the model.
	 */
	public void notifyChanged(Notification notification) {
		int featureId = notification.getFeatureID(VOCLModelPackage.class);
		switch (featureId) {
			case VOCLModelPackage.CONSTRAINT :
				refreshVisuals();
				refreshChildren();
				break;
			default :
				refreshVisuals();
				refreshChildren();
		}
	}

	/**
	 * @see org.eclipse.gef.editparts.AbstractEditPart#refreshVisuals()
	 */
	protected void refreshVisuals() {
		// contextstring, which is displayed
		String context_string = "";

		//current figure and constraint
		ConstraintFigure figure = getConstraintFigure();
		Constraint constraint = getConstraint();

		//Context which belongs to the Constraint
		Context c = (Context) constraint.getConstraintContext();
		context_string += "context ";
		if (!c.getName().equals("")) {
			context_string += c.getName() + " : ";
		}
		context_string += c.getType() + "  ";
		context_string += constraint.getKind().toString().toLowerCase() + " ";

		//no name is given
		if (!constraint.getName().equals("")) {
			context_string += constraint.getName() + " ";
		}
		context_string += ":";

		//insert created String in the figure
		figure.setContext(context_string);

		// Graphical properties
		Dimension size = null;
		if(getConstraint().getHeight() == 0 || getConstraint().getWidth() == 0){
			size = new Dimension(503,400);
		}else{
			size = new Dimension(getConstraint().getWidth(),getConstraint().getHeight());
			getConstraintFigure().setPreferredSize(getConstraint().getWidth(),getConstraint().getHeight());
		}
		Point loc = new Point(getConstraint().getX(), getConstraint().getY());
	
		Rectangle r = new Rectangle(loc, size);
		((GraphicalEditPart) getParent()).setLayoutConstraint(
			this,
			getFigure(),
			r);

	}

	/**
	 * @see org.eclipse.emf.common.notify.Adapter#isAdapterForType(java.lang.Object)
	 */
	public boolean isAdapterForType(Object type) {
		return type.equals(getModel().getClass());
	}

	/**
	 * @see org.eclipse.core.runtime.IAdaptable#getAdapter(java.lang.Class)
	 */
	public Object getAdapter(Class key) {
		if (IPropertySource.class == key)
			return getPropertySource();
		return super.getAdapter(key);
	}

	/**
	 * @see com.ibm.itso.sal330r.gefdemo.edit.NetworkElementEditPart#getPropertySource()
	 */
	protected IPropertySource getPropertySource() {
		if (propertySource == null)
			propertySource = new EObjectPropertySource(getConstraint());
		return propertySource;
	}
	/**
	 * Method getShell
	 * @return
	 */
	public Shell getShell() {
		return shell;
	}

	/**
	 * Method setShell
	 * @param shell
	 */
	public void setShell(Shell shell) {
		this.shell = shell;
	}

	/**
	 * @see org.eclipse.gef.NodeEditPart#getSourceConnectionAnchor(org.eclipse.gef.ConnectionEditPart)
	 */
	public ConnectionAnchor getSourceConnectionAnchor(ConnectionEditPart connection) {
		return null;
	}

	/**
	 * @see org.eclipse.gef.NodeEditPart#getTargetConnectionAnchor(org.eclipse.gef.ConnectionEditPart)
	 */
	public ConnectionAnchor getTargetConnectionAnchor(ConnectionEditPart connection) {

		return null;
	}

	/**
	 * @see org.eclipse.gef.NodeEditPart#getSourceConnectionAnchor(org.eclipse.gef.Request)
	 */
	public ConnectionAnchor getSourceConnectionAnchor(Request request) {
		return null;
	}

	/**
	 * @see org.eclipse.gef.editparts.AbstractEditPart#getModelChildren()
	 */
	protected List getModelChildren() {
		Vector children = new Vector();
		
		if (getConstraint().getVOCLExpression() != null)
			children.add(getConstraint().getVOCLExpression());
		if (getConstraint().getCondition() != null) {
			children.add(getConstraint().getCondition());
		}
		return children;
	}

}
