// File: UMLClassFigureTest.java
package vocleditor.figures;
import java.util.Vector;

import org.eclipse.draw2d.BendpointConnectionRouter;
import org.eclipse.draw2d.ChopboxAnchor;
import org.eclipse.draw2d.ConnectionEndpointLocator;
import org.eclipse.draw2d.Figure;
import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.LightweightSystem;
import org.eclipse.draw2d.PolygonDecoration;
import org.eclipse.draw2d.PolylineConnection;
import org.eclipse.draw2d.PolylineDecoration;
import org.eclipse.draw2d.RelativeBendpoint;
import org.eclipse.draw2d.XYLayout;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.PointList;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;

/**
 * A test class to display a UMLFigure
 */
public class UMLClassFigureTest {
	
	/**
	 * Creates a PolylineConnection between the given source and target figure.
	 * 
	 * @param source The source figure.
	 * @param target The target figure.
	 * @return The polyline connection.
	 */
	static PolylineConnection createConnection(UMLClassFigure source, 
		                                       UMLClassFigure target) {
		PolylineConnection c = new PolylineConnection();
		ChopboxAnchor sourceAnchor = new ChopboxAnchor(source);
		ChopboxAnchor targetAnchor = new ChopboxAnchor(target);
		c.setSourceAnchor(sourceAnchor);
		c.setTargetAnchor(targetAnchor);
		return c;
	}
	
	/**
	 * Transforms the given polyline connection into a composition by adding
	 * a target decoration.
	 *  
	 * @param c The polyline connection.
	 */
	static void makeComposition(PolylineConnection c) {
		PolygonDecoration decoration = new PolygonDecoration();
		PointList decorationPointList = new PointList();
		decorationPointList.addPoint(0, 0);
		decorationPointList.addPoint(-2, 2);
		decorationPointList.addPoint(-4, 0);
		decorationPointList.addPoint(-2, -2);
		decoration.setTemplate(decorationPointList);
		c.setSourceDecoration(decoration);
	}
	
	/**
	 * Makes the target of the given polyline connection navigable by adding
	 * an arrow at the connection end that corresponds to the target.
	 * 
	 * @param c The polyline connection.
	 */
	static void makeTargetNavigable(PolylineConnection c) {
		PolylineDecoration arrow = new PolylineDecoration();
		PointList arrowPointList = new PointList();
		arrowPointList.addPoint(-2, 2);
		arrowPointList.addPoint(0, 0);
		arrowPointList.addPoint(-2, -2);
		arrow.setTemplate(arrowPointList);
		c.setTargetDecoration(arrow);
	}
	
	/**
	 * Adds a label to the given polyline connection. 
	 * 
	 * @param label The label.
	 * @param c The polyline connection.
	 * @param targetLabel If true, the connection end that corresponds to the 
	 *     target will be labeled, otherwise the connection end that corresponds
	 *     to the source will be labeled.
	 * @param d A dimension object.
	 */
	static void createConnectionLabel(String label, 
	                                  PolylineConnection c,
	                                  boolean            targetLabel,
	                                  Dimension          d) {
		ConnectionEndpointLocator endpointLocatorNode =
			new ConnectionEndpointLocator(c, targetLabel);
		endpointLocatorNode.setVDistance(d.width);
		endpointLocatorNode.setUDistance(d.height);
		c.add(new Label(label), endpointLocatorNode);
	}
	
	/**
	 * Main.
	 */
	public static void main(String args[]) {
		Display d = new Display();
		final Shell shell = new Shell(d);
		shell.setSize(500, 400);
		shell.setText("UMLClassFigure Test");
		LightweightSystem lws = new LightweightSystem(shell);
		Figure contents = new Figure();
		XYLayout contentsLayout = new XYLayout();
		contents.setLayoutManager(contentsLayout);

		Font classFont = //new Font(null, "Arial", 12, SWT.BOLD);
			new Font(null, "Sans", 12, SWT.BOLD);
		
		/* 
		 * Create class labels 
		 */
		Label networkLabel = new Label("Network", new Image(d,
			UMLClassFigureTest.class.getResourceAsStream("class_obj.gif"))); 
		networkLabel.setFont(classFont);
		Label nodeLabel = new Label("Node", new Image(d,
			UMLClassFigureTest.class.getResourceAsStream("class_obj.gif"))); 
		nodeLabel.setFont(classFont);
		Label linkLabel = new Label("Link", new Image(d,
			UMLClassFigureTest.class.getResourceAsStream("class_obj.gif"))); 
		linkLabel.setFont(classFont);

		/* 
		 * Create class figures, define their positions on the canvas, and add 
		 * attributes 
		 */
		final UMLClassFigure networkFigure = new UMLClassFigure(networkLabel.getText());
		final UMLClassFigure nodeFigure = new UMLClassFigure(nodeLabel.getText());
		final UMLClassFigure linkFigure = new UMLClassFigure(linkLabel.getText());
		
		// Network's attributes
		Label nameAttrib = new Label("name: EString", new Image(d, 
			UMLClassFigure.class.getResourceAsStream("field_private_obj.gif")));
		networkFigure.getAttributesCompartment().add(nameAttrib);
		
		// Node's attributes			
		Label xAttrib = new Label("x: EInt", new Image(d, 
			UMLClassFigure.class.getResourceAsStream("field_private_obj.gif")));
		Label yAttrib = new Label("y: EInt", new Image(d, 
			UMLClassFigure.class.getResourceAsStream("field_private_obj.gif")));
		nodeFigure.getAttributesCompartment().add(xAttrib);
		nodeFigure.getAttributesCompartment().add(yAttrib);
		
		// Define positions
		contentsLayout.setConstraint(networkFigure, 
			new Rectangle(200, 50, -1, -1));
		contentsLayout.setConstraint(nodeFigure,
			new Rectangle(350, 223, -1, -1));
		contentsLayout.setConstraint(linkFigure,
			new Rectangle(100, 240, -1, -1));

		/* 
		 * Create the connections 
		 */
		PolylineConnection networkNodeConnection = 
			createConnection(networkFigure, nodeFigure);
		makeComposition(networkNodeConnection);
		PolylineConnection networkLinkConnection =
			createConnection(networkFigure, linkFigure);
		makeComposition(networkLinkConnection);
		PolylineConnection strictNodeLinkConnection =
			createConnection(nodeFigure, linkFigure);
		makeTargetNavigable(strictNodeLinkConnection);
		PolylineConnection otherNodeLinkConnection =
			createConnection(nodeFigure, linkFigure);
		makeTargetNavigable(otherNodeLinkConnection);
		
		// Let 'otherNodeLinkConnection' be such that it does not overlaps
		// 'strictNodeLinkConnection'
		RelativeBendpoint p1 = new RelativeBendpoint(otherNodeLinkConnection);
		p1.setRelativeDimensions(new Dimension(0, 50), new Dimension());
		p1.setWeight(0.0f);
		RelativeBendpoint p2 = new RelativeBendpoint(otherNodeLinkConnection);
		p2.setRelativeDimensions(new Dimension(), new Dimension(0, 50));
		p2.setWeight(1.0f);
		Vector bendPoints = new Vector();
		bendPoints.add(p1);
		bendPoints.add(p2);
		BendpointConnectionRouter router = new BendpointConnectionRouter();
		router.setConstraint(otherNodeLinkConnection, bendPoints);
		otherNodeLinkConnection.setConnectionRouter(router);
		
		
		/* 
		 * Label the connections 
		 */
		createConnectionLabel("Network", networkNodeConnection,
			false, new Dimension(20, 10));	
		createConnectionLabel("1", networkNodeConnection,
			false, new Dimension(-5, 25));	 
		 
		createConnectionLabel("Nodes", networkNodeConnection,
			true, new Dimension(-10, 0));	
		createConnectionLabel("0..*", networkNodeConnection,
			true, new Dimension(5, 8));	 
		
		createConnectionLabel("Network", networkLinkConnection,
			false, new Dimension(-15, 10));	
		createConnectionLabel("1", networkLinkConnection,
			false, new Dimension(5, 25));	
		 
		createConnectionLabel("Links", networkLinkConnection,
			true, new Dimension(-2, 10));	
		createConnectionLabel("0..*", networkLinkConnection,
			true, new Dimension(12, 10));	
		 
		createConnectionLabel("upstreamLinks", strictNodeLinkConnection,
			true, new Dimension(2, 20));	
		createConnectionLabel("0..*", strictNodeLinkConnection,
			true, new Dimension(-2, 20));	 

		createConnectionLabel("downstreamLinks", otherNodeLinkConnection,
			true, new Dimension(-5, 20));	
		createConnectionLabel("0..*", otherNodeLinkConnection,
			true, new Dimension(5, 20));	 

		/* 
		 * Add figures to canvas and relax... 
		 */
		contents.add(networkFigure);
		contents.add(nodeFigure);
		contents.add(linkFigure);
		contents.add(networkNodeConnection);
		contents.add(networkLinkConnection);
		contents.add(strictNodeLinkConnection);
		contents.add(otherNodeLinkConnection);
		lws.setContents(contents);
		shell.open();
		while (!shell.isDisposed())
			while (!d.readAndDispatch())
				d.sleep();
	}
}
