/* 
 * $Id: ModelConverter.java,v 1.2 2004/11/09 09:49:37 haensgen Exp $
 *
 * TODO @pre wird bei Methoden-Navigationen noch nicht automatisch erzeugt
 * 
 */
 
package VoclToOclConverter;

import java.util.*;

import ru.novosoft.uml.foundation.core.*;
import ru.novosoft.uml.behavior.collaborations.*;

import de.tuberlin.tfs.ocl.metamodel.expressions.*;
import de.tuberlin.tfs.ocl.metamodel.types.*;
import de.tuberlin.tfs.ocl.metamodel.converter.*;

import VOCLModel.*;

/**
 * Provides methods to convert a given Visual OCL diagram into OCL expressions 
 * for each Visual OCL expression which occurs in the diagram. Visual OCL 
 * diagrams are given by an instance of {@link VOCLModel.VOCLDiagram}. OCL
 * expressions are represented by instances of 
 * {@link de.tuberlin.tfs.ocl.metamodel.expressions.MOclConstraint}.
 * Additionally, OCL expressions can be converted into their textual 
 * representations, called OCL strings.
 * <p>
 * An example of a session where a given Visual OCL diagram is converted into
 * the corresponding OCL strings could be as follows.
 * <pre>
 *     VOCLDiagram root = ...; // Root of the Visual OCL diagram
 *     ModelConverter converter = new ModelConverter(root);
 *     String[] oclStrings = null;
 *     try {
 *         oclStrings = converter.convertToOclStrings();
 *         // do something with oclStrings
 *     } catch (ModelConverterException ex) {
 *         // catch exception
 *     }
 * </pre>
 * 
 * @see #convertToOclModels()
 * @see #convertToOclStrings()
 */
public class ModelConverter {

	// The root element of the EMF model instance to convert.
	private VOCLDiagram root = null;
	
	// To collect bad elements during conversion.
	private BadElementCollector badElements = null;
	
	// Hashtable to remember types.
	private Hashtable types = null; 
	
	// Hashtable to remember operations.
	private Hashtable operations = null;
	
	// Hashtable to remember variable declarations.
	private Hashtable variableDeclarations = null;
	
	/**
	 * Creates a new ModelConverter instance for the EMF model instance whose
	 * root is given.
	 * 
	 * @param root
	 *     The root of the EMF model instance.
	 * 
	 * @exception NullPointerException
	 *     If <tt>root</tt> is <tt>null</tt>.
	 * @exception InvalidModelException 
	 *     If the model is invalid. 
	 */
	public ModelConverter(VOCLDiagram root) throws InvalidModelException {
		this.root = root;
		
		// Validate model first.
		ModelValidator validator = new ModelValidator(root);
		if (!validator.isValid())
			throw new InvalidModelException(
				validator.getInvalidElementStates());
			
		badElements = new BadElementCollector(root, 100);
		
		// Create table for types and fill it with standard types.	
		types = new Hashtable();
		
		MClassifier booleanType = new MClassifierImpl();
		booleanType.setName("Boolean");
		types.put("Boolean", booleanType);
		
		MClassifier integerType = new MClassifierImpl();
		integerType.setName("Integer");
		types.put("Integer", integerType);
		
		MClassifier realType = new MClassifierImpl();
		realType.setName("Real");
		types.put("Real", realType);
		
		MClassifier stringType = new MClassifierImpl();
		stringType.setName("String");
		types.put("String", stringType);
		
		// Create table for operations and fill it with standard operations.
		operations = new Hashtable();
		
		MOperation equalityOperation = new MOperationImpl();
		equalityOperation.setName("=");
		operations.put("=", equalityOperation);
		
		MOperation andOperation = new MOperationImpl();
		andOperation.setName("and");
		operations.put("and", andOperation);		
		
		MOperation orOperation = new MOperationImpl();
		orOperation.setName("or");
		operations.put("or", orOperation);
		
		MOperation xorOperation = new MOperationImpl();
		xorOperation.setName("xor");
		operations.put("xor", xorOperation);
		
		MOperation impliesOperation = new MOperationImpl();
		impliesOperation.setName("implies");
		operations.put("implies", impliesOperation);
		
		// Create table for variables.
		variableDeclarations = new Hashtable();
	}
	
	/**
	 * Returns a list of strings which represent the OCL expressions which 
	 * occur in the Visual OCL diagram.
	 *
	 * @throws ModelConverterException
	 *     If an error occured.
	 */
	public String[] convertToOclStrings() throws ModelConverterException {
		MOclConstraint[] constraint = convertToOclModels();
		
		Vector strings = new Vector();
		for (int i = 0; i < constraint.length; i++) {
			try {
				strings.add(constraint[i].convert2OclString());
			} catch (NoSuchAttributeException ex) {
				// Should not occur. Ignore and drop a debug message.
				System.out.println("[Warning] Unable to convert " +
					"MOclConstraint: " + ex.getLocalizedMessage());
			}
		}
		
		return (String[])strings.toArray(new String[0]);
	}

	/**
	 * Converts the constraints in the Visual OCL model into OCL meta model 
	 * instances.
	 * 
	 * @return
	 *     An array of OCL meta model instances (whose root is given by a 
	 *     MOclConstraint instance).
	 * @throws ModelConverterException
	 *     If an error occurs.
	 */
	public MOclConstraint[] convertToOclModels() throws ModelConverterException {
		System.out.println("[Debug] Converting...");
		
		Vector models = new Vector();
		for (Iterator it = root.getConstraint().iterator(); it.hasNext(); ) {
			Constraint constraint = (Constraint)it.next();
			try {
				MOclConstraint model = convertConstraint(constraint);
				if (model != null)
					models.add(model);
				else {
					// Should not happen. Print message for debug purposes.
					System.out.println("[Warning] convertToOclModel(" + 
						constraint + ") == null");
				}
			} catch (ModelConverterException ex) {
				// Ignore silently.
			}
		}
		
		System.out.println("[Debug] Converted successfully.");
	
		return (MOclConstraint[])models.toArray(new MOclConstraint[0]);
	}

	/**
	 * Converts the given constraint into an OCL meta model instance.
	 * 
	 * @param constraint
	 *     The constraint to convert.
	 * @return 
	 *     A MOclConstraint which is the root of the created meta model
	 *     instance.
	 * 
	 * @exception ModelConverterException
	 *     If an error occured while converting.
	 */
	public MOclConstraint convertConstraint(Constraint constraint) 
		throws ModelConverterException {
		
		MOclConstraint oclConstraint = new MOclConstraintImpl();
		
		// Set name.
		if (constraint.getName() != null)
			oclConstraint.setName(constraint.getName());
		
		// Set kind.
		switch (constraint.getKind().getValue()) {
		case ConstraintKind.PRE:
			oclConstraint.setKind(MOclConstraintKind.PRE);
			break;
		case ConstraintKind.POST:
			oclConstraint.setKind(MOclConstraintKind.POST);
			break;
		default:
			oclConstraint.setKind(MOclConstraintKind.INV);
		}
		
		// Set constrained element.
		Context context = constraint.getConstraintContext();
		oclConstraint.setConstrainedElement(convertContext(context));
		
		// Set body.
		VOCLExpression exp = constraint.getVOCLExpression();
		if (exp != null) {
			MOclExpression body = 
				convertVoclExpression(exp, context, new Vector());
			oclConstraint.setBody(body);
		} 
		
		return oclConstraint;
	}
	
	/**
	 * Converts a given Context instance into an OCL meta model element which 
	 * can be used as the constrained element attribute for the OCL constraint.
	 * 
	 * @param context
	 *     The context to consider.
	 * @return 
	 *     A MModelElement instant which is the described element.
	 * 
	 * @exception ModelConverterException
	 *     If an error occured while converting.
	 */
	public MModelElement convertContext(Context context) 
		throws ModelConverterException {	
			
		MModelElement element = null;
			
		MClassifier type = getType(context.getType());
		String name = context.getName();
		
		if (context.isIsClassifier()) {
			List varDecls = context.getVariableDeclaration();
			if (varDecls != null && varDecls.size() > 0) {
				VariableDeclaration varDecl = (VariableDeclaration)varDecls.get(0);

				// (Todo) Wie setzt man den Variablennamen in den Constraint-Kopf???
				// Work-a-round: "<variable>:<Typ>" als Typ setzen.
				type = getType(varDecl.getVarName() + ":" + varDecl.getType());
			}
			element = type;
		} else {
			// Get name of classifier and name of method.
			int splitPos = name.indexOf(':');
			int paramBegin = name.indexOf('(');
			if (paramBegin == -1)
				paramBegin = name.length() - 1;
			String classifierName = name.substring(0, splitPos);
			String methodName = name.substring(splitPos + 2, paramBegin);
			
			MOperationCallExp opCallExp = new MOperationCallExpImpl();

			MOperation operation = new MOperationImpl();
			operation.setName(classifierName + "::" + methodName);
			opCallExp.setReferredOperation(operation);
			
			List methodVars = getVariablesFromMethod(name);
			for (Iterator varIt = methodVars.iterator(); 
				 varIt.hasNext(); ) {
				String var = (String)varIt.next();
				String varName = getVarName(var);
				String varType = getVarType(var);
				
				MAttributeCallExp attrCallExp = new MAttributeCallExpImpl();
				
				MClassifier attrType = getType(varType);
				attrCallExp.setType(attrType);
				
				MAttribute attr = new MAttributeImpl();
				attr.setName(varName);
				attr.setType(attrType);
				
				attrCallExp.setReferredAttribute(attr);
				opCallExp.addArguments(attrCallExp); 
			}
			
			MClassifier returnType = getType(context.getType());
			opCallExp.setType(returnType);
			
			element = opCallExp;
		}
		
		return element;
	}
	
	/**
	 * Converts a given Visual OCL expression into an OCL expression.
	 * 
	 * @param exp
	 *     The expression to convert.
	 * @param context
	 *     The context in which the VOCL expression is contained in.
	 * @param boundedVariables
	 *     A list of names of variables which are bounded, for instance via 
	 *     "let".
	 * @return
	 *     An OCL expression given as MOclExpression instance.
	 * 
	 * @throws ModelConverterException
	 *     If an error occurs.
	 */
	public MOclExpression convertVoclExpression(VOCLExpression exp,
	                                            Context context,
	                                            List boundedVariables) 
		throws ModelConverterException {
			
		if (exp == null)
			return null;
			
		// The OCL expression to return.
		MOclExpression oclExp = null;
		
		if (exp instanceof NavExp)
			oclExp = 
				convertNavExp(
					(NavExp)exp, context, boundedVariables);	
		else if (exp instanceof BoolExp) {
			BoolExp boolExp = (BoolExp)exp;
			
			MOperationCallExp opCallExp = new MOperationCallExpImpl();
			
			MOclExpression firstOp =
				convertVoclExpression(
					boolExp.getFirstOp(), context, boundedVariables); 
			MOclExpression secondOp =
				convertVoclExpression(
					boolExp.getSecondOp(), context, boundedVariables); 
					
			if (firstOp != null && secondOp != null) {
				opCallExp.addArguments(firstOp);
				opCallExp.addArguments(secondOp);
			} else
				return null;
			
			opCallExp.setType(getType("Boolean"));
			switch (boolExp.getKind().getValue()) {
			case BoolExpKind.AND:
				opCallExp.setReferredOperation(getOperation("and"));
				break;
			case BoolExpKind.OR:
				opCallExp.setReferredOperation(getOperation("or"));
				break;
			case BoolExpKind.XOR:
				opCallExp.setReferredOperation(getOperation("xor"));
				break;
			case BoolExpKind.IMPLIES:
				opCallExp.setReferredOperation(getOperation("implies"));
				break;
			}
			
			oclExp = opCallExp;
		} else if (exp instanceof IfExp) {
			IfExp ifExp = (IfExp)exp;
			
			MIfExp oclIfExp = new MIfExpImpl();
			
			MOclExpression condition =
				convertVoclExpression(
					ifExp.getIfCondition(), context, boundedVariables);
			MOclExpression thenExp =
				convertVoclExpression(
					ifExp.getThenExp(), context, boundedVariables);
			MOclExpression elseExp =
				convertVoclExpression(
					ifExp.getElseExp(), context, boundedVariables);
					
			if (condition != null && thenExp != null && elseExp != null) {
				oclIfExp.setCondition(condition);
				oclIfExp.setTheExpression(thenExp);
				oclIfExp.setElseExpression(elseExp);
				oclIfExp.setType(thenExp.getType());
			} else
				return null;
			
			oclExp = oclIfExp;
		} else if (exp instanceof LetExp) {
			LetExp letExp = (LetExp)exp;
			VariableDeclaration letVar = letExp.getVariableDeclaration();
			List newBoundedVariables = new Vector(boundedVariables);
			newBoundedVariables.add(letVar.getVarName());
			
			MLetExp oclLetExp = new MLetExpImpl();
			
			MOclExpression def =
				convertVoclExpression(
					letExp.getDef(), context, newBoundedVariables);
			boolean convertedDef = false;
			if (!letVar.getType().equals("Boolean")) {
				if (def instanceof MOperationCallExp) {
					MOperationCallExp opCallExp = (MOperationCallExp)def;
					List args = opCallExp.getArguments();
					if (args.size() >= 2) {
						if (args.get(1) instanceof MVariableExp) {
							MVariableDeclaration decl = 
								((MVariableExp)args.get(1)).getReferredVariable();
							String name = decl.getVarname();
							if (name.equals(letVar.getVarName())) {
								def = (MOclExpression)args.get(0);
								convertedDef = true;
							}
						}
					}
				}
			} else
				convertedDef = true;
					
			MOclExpression in =
				convertVoclExpression(
					letExp.getInExp(), context, newBoundedVariables);
			
			if (convertedDef && in != null) {
				MVariableDeclaration varDecl = new MVariableDeclarationImpl();
				varDecl.setType(getType(letVar.getType()));
				varDecl.setVarname(letVar.getVarName());
				varDecl.setInitExpression(def);
				oclLetExp.setVariable(varDecl);
				oclLetExp.setIn(in);
				oclLetExp.setType(in.getType());
			} else
				return null;
					
			oclExp = oclLetExp;
		}
		
		if (exp.isIsNegated()) {
			MOperationCallExp notExp = new MOperationCallExpImpl();
			notExp.setType(getType("Boolean"));
			notExp.setReferredOperation(getOperation("not"));
			notExp.addArguments(oclExp);
			oclExp = notExp;
		}
		
		return oclExp;
	}
	
	/**
	 * Converts a given navigation expression with respect to the specified
	 * list of start variables and bounded variables.
	 * 
	 * @param navExp
	 *     The navigation expression to convert.
	 * @param context
	 *     The context, in which the navExp is considered.
	 * @param boundedVariables
	 *     A list of names of variables which are bounded, e.g. by a let 
	 *     expression.
	 * @return
	 *     An OCL expression given as MOclExpression instance.
	 * 
	 * @throws ModelConverterException
	 *     If an error occurs.
	 */
	public MOclExpression convertNavExp(NavExp navExp, 
	                                    Context context,
	                                    List boundedVariablesList) 
		throws ModelConverterException {	
		
		System.out.println("[Debug] Convert navigation expression...");
		
		MOclExpression oclExp = null;
		
		// Get condition, if there is any
		Condition condition = navExp.getCondition();
		if (condition == null) {
			Constraint constraint = navExp.getConstraint();
			if (constraint != null)
				condition = constraint.getCondition();
		}
		
		if (condition != null && condition.getRight() == null) {
			VariableDeclaration leftVar = condition.getLeft();
			String leftVarName = leftVar.getVarName();
			
			boolean isBounded = false;
			for (Iterator it = boundedVariablesList.iterator(); 
				 !isBounded && it.hasNext(); ) {
				String varName = (String)it.next();
				if (varName.equals(leftVarName))
					isBounded = true;
			}
			if (isBounded) {
				 MOperationCallExp exp = new MOperationCallExpImpl();
				 exp.setReferredOperation(getOperation(condition.getOp()));
				 exp.setType(getType("Boolean"));
				 
				 // Left side.
				 MVariableDeclaration varDecl =
					 getVariableDeclaration(leftVarName);
				 varDecl.setType(getType(leftVar.getType()));
				 MVariableExp varExp = new MVariableExpImpl();
				 varExp.setType(varDecl.getType());
				 varExp.setReferredVariable(varDecl);
				 exp.addArguments(varExp);
	
				 // Right side.
				 MLiteralExp valueExp = null;
				 String type = leftVar.getType();
				 if (type.equals("Boolean"))
					valueExp = new MBooleanLiteralExpImpl();
				 else if (type.equals("Integer"))
					valueExp = new MIntegerLiteralExpImpl();
				else if (type.equals("Real"))
					valueExp = new MRealLiteralExpImpl();
				else if (type.equals("String"))
					valueExp = new MStringLiteralExpImpl();
				else
					valueExp = new MLiteralExpImpl();
				valueExp.setType(getType(type));
				valueExp.setName(condition.getValue());
				exp.addArguments(valueExp);
		
				 oclExp = concatenate(oclExp, exp);
			}
		}
		
		// Bind start variables.
		Vector boundedVariables = new Vector(boundedVariablesList);
		List startVars = getStartVariables(context.getConstraint());
		if (startVars != null && startVars.size() > 0)
			for (Iterator it = startVars.iterator(); it.hasNext(); ) 
				boundedVariables.add(it.next());
		
		// Get all navigation paths in the model.
		Vector paths = new Vector();
		List startNodes = getStartNodes(navExp, startVars);
		System.out.println("[Debug] Found start nodes: " + startNodes);
		for (Iterator nodeIt = startNodes.iterator(); nodeIt.hasNext(); ) {
			ClassifierNode node = (ClassifierNode)nodeIt.next();
			paths.addAll(getPaths(node));
		}
		System.out.println("[Debug] Found paths: " + paths);
		
		// Generate a list of converted paths. These paths may have variable
		// expressions on their right sides which we replace later.
		Stack expressions = new Stack();
		for (ListIterator it = paths.listIterator(); it.hasNext(); ) {
			NavigationPath path = (NavigationPath)it.next();
			if (path.isRoundTrip()) {
				MOclExpression exp = convertRoundTrip(path);
				if (exp != null)
					expressions.add(exp);
			}
			if (path.isAttributeCall()) {
				MOclExpression[] exp = convertAttributeCall(path);
				if (exp != null) 
					for (int i = 0; i < exp.length; i++)
						expressions.add(exp[i]);
			}
			if (path.isEmptyOperation()) {
				MOclExpression exp = convertEmptyOperation(path);
				if (exp != null)
					expressions.add(exp);
			}
			if (path.isNotEmptyOperation()) {
				MOclExpression exp = convertNotEmptyOperation(path);
				if (exp != null)
					expressions.add(exp);
			}
			if (path.isSizeOperation()) {
				MOclExpression exp = convertSizeOperation(path);
				if (exp != null)
					expressions.add(exp);
			}
			if (path.isSumOperation()) {
				MOclExpression exp = convertSumOperation(path);
				if (exp != null)
					expressions.add(exp);
			}
			if (path.isMethodCall()) {
				MOclExpression exp = convertMethodCall(path);
				if (exp != null)
					expressions.add(exp);
			}
			
			// Search for equivalent paths
			for (Iterator eqIt = paths.listIterator(it.nextIndex()); 
			     eqIt.hasNext(); ) {
				NavigationPath curPath = (NavigationPath)eqIt.next();
				if (path.isEquivalent(curPath)) {
					MOclExpression exp = convertEquivalentPaths(path, curPath);
					if (exp != null)
						expressions.add(exp);
				}
			}
		}
		
		// Build OCL expression.
		while (!expressions.isEmpty()) {
			MOclExpression exp = (MOclExpression)expressions.pop();
			
			if (exp instanceof MOperationCallExp) {
				MOperationCallExp opCallExp = (MOperationCallExp)exp;
				List args = opCallExp.getArguments();
				if (args.size() >= 2) {
					if (args.get(1) instanceof MVariableExp) {
						MVariableDeclaration decl = 
							((MVariableExp)args.get(1)).getReferredVariable();
						String name = decl.getVarname();

						boolean success = false;
						MOclExpression[] exps = 
							processCondition(
								condition,
								(MOclExpression)args.get(0),
								name,
								expressions,
								boundedVariables);
						if (exps != null && exps.length > 0) {
							success = true;
							for (int i = 0; i < exps.length - 1; i++) 
								oclExp = concatenate(oclExp, exps[i]);
							exp = exps[exps.length - 1];
						}

						if (!success) {
							boolean isBounded = false;
							for (Iterator it = boundedVariables.iterator(); 
						    	 !isBounded && it.hasNext(); ) {
						    	String varName = (String)it.next();
								if (varName.equals(name))
									isBounded = true;
							}

							if (!isBounded) 
								exp = null;
						}
					}
				}
			}
			
			if (exp != null)
				oclExp = concatenate(oclExp, exp);
		}
		
		return oclExp;
	}
	
	/**
	 *
	 */
	protected MOclExpression[] processCondition(Condition condition,
	                                            MOclExpression firstArg,
	                                            String varName,
	                                            List expressions,
	                                            List boundedVariables) {	
	    if (condition == null)
	    	return null;
	                                            	
	    List oclExps = new Vector();
	    
	    String leftName = condition.getLeft().getVarName();
	    
		if (condition.getRight() == null) { // right side is literal
			if (leftName.equals(varName)) {
				// Root operation
				MOperationCallExp exp = new MOperationCallExpImpl();
				exp.setReferredOperation(getOperation(condition.getOp()));
				exp.setType(getType("Boolean"));
				
				// Add left side
				exp.addArguments(firstArg);
				
				// Add right side
				MLiteralExp valueExp = null;
				String type = firstArg.getType().getName();
				if (type.equals("Boolean"))
					valueExp = new MBooleanLiteralExpImpl();
				else if (type.equals("Integer"))
					valueExp = new MIntegerLiteralExpImpl();
				else if (type.equals("Real"))
					valueExp = new MRealLiteralExpImpl();
				else if (type.equals("String"))
					valueExp = new MStringLiteralExpImpl();
				else
					valueExp = new MLiteralExpImpl();
				valueExp.setType(firstArg.getType());
				valueExp.setName(condition.getValue());
				exp.addArguments(valueExp);
				oclExps.add(exp);
			}
		} else { // variables on both sides
			String rightName = condition.getRight().getVarName();
			
			String searchName = null;
			if (varName.equals(leftName))
				searchName = rightName;
			else if (varName.equals(rightName))
				searchName = leftName;
			
			if (searchName != null) {
				boolean foundExp = false;
				Vector copy = new Vector(expressions);
				for (Iterator it = copy.iterator(); it.hasNext(); ) {
					MOclExpression curExp = (MOclExpression)it.next();
					if (curExp instanceof MOperationCallExp) {
						List args = ((MOperationCallExp)curExp).getArguments();
						if (args.size() < 2) continue;
						if (args.get(1) instanceof MVariableExp) {
							MVariableDeclaration decl = 
								((MVariableExp)args.get(1)).getReferredVariable();
							if (decl.getVarname().equals(searchName)) {
								MOclExpression secondArg = (MOclExpression)args.get(0);
												
								// Root operation
								MOperationCallExp exp = new MOperationCallExpImpl();
								exp.setReferredOperation(getOperation(condition.getOp()));
								exp.setType(getType("Boolean"));

								if (varName.equals(leftName)) {
									exp.addArguments(firstArg);
									exp.addArguments(secondArg);
								} else {
									exp.addArguments(secondArg);
									exp.addArguments(firstArg);
								}

								oclExps.add(exp);
				
								expressions.remove(curExp);
								
								foundExp = true;
							}
						}
					}
				}
				
				if (!foundExp) {
					boolean isBounded = false;
					for (Iterator it = boundedVariables.iterator(); 
						 !isBounded && it.hasNext(); ) {
						String name = (String)it.next();
						if (searchName.equals(name))
							isBounded = true;
					}
				
					if (isBounded) {
						MOperationCallExp exp = new MOperationCallExpImpl();
						exp.setReferredOperation(getOperation(condition.getOp()));
						exp.setType(getType("Boolean"));
					
						if (varName.equals(leftName))
							exp.addArguments(firstArg);
					
						// Add right side
						MVariableDeclaration varDecl =
							getVariableDeclaration(searchName);
						varDecl.setType(firstArg.getType());
						MVariableExp varExp = new MVariableExpImpl();
						varExp.setType(varDecl.getType());
						varExp.setReferredVariable(varDecl);
						exp.addArguments(varExp);
						
						if (varName.equals(rightName))
							exp.addArguments(firstArg);
							
						oclExps.add(exp);
					}
				}
			}
		}
		
		return (MOclExpression[])oclExps.toArray(new MOclExpression[0]);
	}
	
	/**
	 * Returns the OCL type with the given name. If it does not exist, it will 
	 * be created.
	 * 
	 * @param name
	 *     The name of the type to return.
	 */
	protected MClassifier getType(String name) {
		MClassifier type = (MClassifier)types.get(name);
		if (type == null) {
			type = new MClassifierImpl();
			type.setName(name);
			types.put(name, type);
		}
		return type;
	}
	
	/**
	 * Returns the OCL operation with the given name. If it does not exist, it 
	 * will be created.
	 * 
	 * @param name
	 *     The name of the operation to return.
	 */
	protected MOperation getOperation(String name) {
		MOperation operation = (MOperation)operations.get(name);
		if (operation == null) {
			operation = new MOperationImpl();
			operation.setName(name);
			operations.put(name, operation);
		}
		return operation;
	}
	
	/**
	 * Returns the OCL variable declaration with the given name. If it does 
	 * not exist, it will be created. Note that the type will not be set.
	 * 
	 * @param name
	 *     The name of the variable declaration to return.
	 */
	protected MVariableDeclaration getVariableDeclaration(String name) {
		MVariableDeclaration varDecl = 
			(MVariableDeclaration)variableDeclarations.get(name);
		if (varDecl == null) {
			varDecl = new MVariableDeclarationImpl();
			varDecl.setVarname(name);
			variableDeclarations.put(name, varDecl);
		}
		return varDecl;
	}
	
	/**
	 * Returns all variables of classifier nodes where navigation in navigation
	 * expressions defined in the given constraint starts.
	 * 
	 * @param constraint
	 *     The constraint to return the variables for.
	 */
	protected List getStartVariables(Constraint constraint) {
		Context context = constraint.getConstraintContext();
		String name = context.getName();
		List vars = context.getVariableDeclaration();
		Vector startVars = new Vector();
		
		if (name != null) {
			if (context.isIsClassifier()) {
				if (vars == null || vars.size() == 0)
					startVars.add("self");
				else
					for (Iterator it = vars.iterator(); it.hasNext(); ) {
						VariableDeclaration var = (VariableDeclaration)it.next();
						if (name.equals(var.getVarName())) {
							startVars.add(var.getVarName());
							break;
						}
					}
			} else {
				startVars.add("self"); // for the classifier
				//startVars.add("result"); // for the return value
				List methodVars = getVariablesFromMethod(context.getName());
				for (Iterator it = methodVars.iterator(); it.hasNext(); ) {
					String var = (String)it.next();
					startVars.add(getVarName(var));
				}
			}
		}
		
		return startVars;
	}
	
	/**
	 * Liefert die Startrollen eines Constraints zurueck.
	 * 
	 * @return
	 *     Die Startrollen des angegebenen Constraints oder <tt>null</tt>, falls
	 *     keine existieren.
	 */
	protected static List getStartNodes(NavExp navExp, List startVariables) {
		List nodes = new Vector();
		
		for (Iterator it = navExp.getClassifierNode().iterator(); it.hasNext(); ) {
			ClassifierNode node = (ClassifierNode)it.next();
			VariableDeclaration var = node.getClassifierVar();
			if (var != null) {
				for (Iterator varIt = startVariables.iterator(); varIt.hasNext(); ) {
					if (var.getVarName().equals(varIt.next()))
						nodes.add(node);	
				}
			}
		}
		
		return nodes;
	}
	
	/**
	 * Returns a list of navigation paths which can be found in the Visual OCL 
	 * expression and start at the given classifier node.
	 * 
	 * @param startNode
	 *     The start classifier node.
	 * @return
	 *     A list of NavigationPaths starting at startNode.
	 */
	protected static List getPaths(ClassifierNode startNode) {
		Vector paths = new Vector(); // paths to return
		Stack  stack = new Stack();  // stack for depth-first search
		
		NavigationPath startPath = new NavigationPath(startNode);
		stack.push(startPath);
		paths.add(startPath);
		
		while (!stack.isEmpty()) {
			NavigationPath path = (NavigationPath)stack.pop();
			ClassifierNode lastNode = path.getLastNode();
			
			// Is last node an end node? Then, path is an acyclic path which 
			// must not be considered further.
			List outgoingAssociations = getOutgoingAssociations(lastNode);
			if (lastNode != startNode && 
			    (outgoingAssociations == null || 
			    outgoingAssociations.size() == 0))
				paths.add(path);
			else {
				for (Iterator it = outgoingAssociations.iterator(); it.hasNext(); ) {
					AssociationRole ar = (AssociationRole)it.next();
					ClassifierNode cr = 
						ar.getAssociationEndRoleTarget().getClassifierNode();
	
					// Check, if the extension of path by ar and cr results in a
					// cycle. If so, and if the first role equals the last one, we
					// found a round trip. Otherwise, if it results in cycle but is
					// not a round trip, we reject the path.
					if (path.contains(cr)) {
						if (cr.equals(startNode)) {
							NavigationPath roundTrip = new NavigationPath(path);
							try {
								roundTrip.add(ar);
							} catch (IncompatibleAssociationRoleException ex) {
								// ignore
							}
							paths.add(roundTrip);
						}
					} else {
						NavigationPath extension = new NavigationPath(path);
						try {
							extension.add(ar);
						} catch (IncompatibleAssociationRoleException ex) {
							// ignore
						}
						stack.push(extension);
					}
				}
			}
		}

		return paths;
	}
	
	/**
	 * Returns an OCL expression for the given path. 
	 * 
	 * @param path
	 *     The path for which to create the expression.
	 * @return
	 *     An OCL expression for the given path.
	 */
	protected MOclExpression convertNavigationPath(NavigationPath path) {
		// Create type for navigation start node.
		ClassifierNode startNode = path.getFirstNode();
		MClassifier startType = getType(startNode.getType());
        MVariableDeclaration startVariableDeclaration = 
        	getVariableDeclaration(startNode.getClassifierVar().getVarName());
        startVariableDeclaration.setType(startType);
		
		// Create expression.
		MVariableExp startExp = new MVariableExpImpl();
		startExp.setType(startType);
		startExp.setReferredVariable(startVariableDeclaration);
		MOclExpression expression = startExp;

		Iterator it = path.iterator();
		ClassifierNode lastNode = (ClassifierNode)it.next(); 
		while (it.hasNext()) {
			AssociationRole associationRole = 
				(AssociationRole)it.next();
			AssociationEndRoleSource sourceEndRole = 
				associationRole.getAssociationEndRoleSource();
			AssociationEndRoleTarget targetEndRole = 
				associationRole.getAssociationEndRoleTarget();
			ClassifierNode classifierNode = 
				(ClassifierNode)it.next();
				
			// Create new AssociationEndCallExp. This will be the new left side
			// then. 
			MAssociationEndCallExp newExpression = new MAssociationEndCallExpImpl();
			newExpression.setSource(expression);
			MClassifier type = getType(classifierNode.getType());
			newExpression.setType(type);
			expression = newExpression;
			
			// Create source and target association ends.
			MAssociationEnd sourceAssociationEnd = new MAssociationEndImpl();
			newExpression.setNavigationSource(sourceAssociationEnd);
			sourceAssociationEnd.setName(sourceEndRole.getName());
			MClassifier sourceAssociationEndType = 
				getType(sourceEndRole.getClassifierNode().getType());
			sourceAssociationEnd.setType(sourceAssociationEndType);
			
			MAssociationEnd targetAssociationEnd = new MAssociationEndImpl();
			newExpression.setReferredAssociationEnd(targetAssociationEnd);
			targetAssociationEnd.setName(targetEndRole.getName());
			
			MClassifier targetAssociationEndType = 
				getType(targetEndRole.getClassifierNode().getType());
			targetAssociationEnd.setType(targetAssociationEndType);
			
			// Create Association, AssociationRole, and AssociationEndRoles
			MAssociation oclAssociation = new MAssociationImpl();
			oclAssociation.addConnection(sourceAssociationEnd);
			oclAssociation.addConnection(targetAssociationEnd);
			
			MAssociationRole oclAssociationRole = new MAssociationRoleImpl();
			oclAssociationRole.setBase(oclAssociation);
			
			MAssociationEndRole oclSourceEndRole = new MAssociationEndRoleImpl();
			oclAssociationRole.addConnection(oclSourceEndRole);
			oclSourceEndRole.setBase(sourceAssociationEnd);
			MClassifierRole oclSourceEndRoleType = new MClassifierRoleImpl();
			oclSourceEndRoleType.setName(lastNode.getType());
			oclSourceEndRole.setType(oclSourceEndRoleType);
			
			MAssociationEndRole oclTargetEndRole = new MAssociationEndRoleImpl();
			oclAssociationRole.addConnection(oclTargetEndRole);
			oclTargetEndRole.setBase(targetAssociationEnd);
			MClassifierRole oclTargetEndRoleType = new MClassifierRoleImpl();
			oclTargetEndRoleType.setName(classifierNode.getType());
			oclTargetEndRole.setType(oclTargetEndRoleType);
			
			lastNode = classifierNode;
		}	
			
		return expression;
	}
	
	/**
	 * Creates an OCL expression for the given round trip.
	 * 
	 * @param path
	 *     A round trip path.
	 * @exception NoRoundTripException 
	 *     If the given path is no round trip.
	 * @return
	 *     The OCL expression which represents the round trip or <tt>null</tt>
	 *     if the navigation path does not represent a round trip.
	 */
	protected MOclExpression convertRoundTrip(NavigationPath path) {
		if (!path.isRoundTrip())
			return null;
		
		// Root operation
		MOperationCallExp expression = new MOperationCallExpImpl();
		expression.setReferredOperation(getOperation("="));
		expression.setType(getType("Boolean"));
		
		// Create type for navigation start role.
		ClassifierNode startNode = path.getFirstNode();
		MClassifier startType = new MClassifierImpl();
		startType.setName(startNode.getType());
        
		// Create variable declaration for start role.
		MVariableDeclaration startVariableDeclaration = 
			new MVariableDeclarationImpl();
		startVariableDeclaration.setVarname(
			startNode.getClassifierVar().getVarName());
		startVariableDeclaration.setType(startType);
		
		// Left hand side
		MOclExpression leftSide = convertNavigationPath(path);
		leftSide.setParentOperation(expression);
		expression.addArguments(leftSide);
		
		// Right hand side
		MVariableExp rightSide = new MVariableExpImpl();
		rightSide.setParentOperation(expression);
		rightSide.setType(startType);
		rightSide.setReferredVariable(startVariableDeclaration);
		expression.addArguments(rightSide);
		
		return expression;
	}
	
	
	/**
	 * Creates an OCL meta model expression for the given equivalent paths.
	 * 
	 * @param path1
	 *     The first navigation path.
	 * @param path2
	 *     The second navigation path.
	 * 
	 * @exception NoEquivalentPathsException
	 *     If the given paths are not equivalent.
	 */
	protected MOclExpression convertEquivalentPaths(NavigationPath path1,
													NavigationPath path2) {
		if (!path1.isEquivalent(path2))
			return null;
				
		// Root operation
		MOperationCallExp expression = new MOperationCallExpImpl();
		expression.setReferredOperation(getOperation("="));
		expression.setType(getType("Boolean"));
		
		// Create the left and right hand side of the operation.
		MOclExpression leftSide = convertNavigationPath(path1);
		leftSide.setParentOperation(expression);
		expression.addArguments(leftSide);
		
		MOclExpression rightSide = convertNavigationPath(path2);
		rightSide.setParentOperation(expression);
		expression.addArguments(rightSide);
		
		return expression;	
	}
	
	
	/**
	 * Converts an attribute call navigation path to the corresponding OCL meta
	 * model expression. For each attribute associated with the last role on the
	 * given path, an own expression will be created. If more than one attribute
	 * exists, all expressions will be concatenated by "and".
	 *  
	 * @param path
	 *     The path which represents the attribute call.
	 * @return
	 *     The OCL expression which corresponds to the attribute call.
	 * 
	 * @exception NoAttributeCallException
	 *     If the given path does not represent an attribute call expression.
	 */
	protected MOclExpression[] convertAttributeCall(NavigationPath path) {
		if (!path.isAttributeCall())
			return null;
		
		Vector exprList = new Vector();
		
		List attributes = ((ClassifierNode)path.getLastNode()).getAttribute();
		for (Iterator it = attributes.iterator(); it.hasNext(); ) {
			Attribute attribute = (Attribute)it.next();
			MOclExpression exp = convertAttributeCall(path, attribute);
			if (exp != null)
				exprList.add(exp);
		}
		
		return (MOclExpression[])exprList.toArray(new MOclExpression[0]);
	}
	
	/**
	 * Converts the given attribute call navigation path (with respect to the
	 * given attribute) to the corresponding OCL meta model expression.
	 * 
	 * @param path
	 *     The navigation path. This must be an attribute call expression.
	 * @param attribute
	 *     The attribute. This must be an attribute associated with the last 
	 *     role of the given path.
	 * 
	 * @exception NoAttributeCallException
	 *     If the given path does not represent an attribute call expression,
	 *     or if the attribute is not an attribute of the last classifier role
	 *     on the path.
	 */
	protected MOclExpression convertAttributeCall(NavigationPath path,
												  Attribute attribute) {
		if (!path.isAttributeCall())
			return null;
		else if (!((ClassifierNode)path.getLastNode()).getAttribute().contains(attribute))
			return null;
				
		// Root operation
		MOperationCallExp expression = new MOperationCallExpImpl();
		expression.setReferredOperation(getOperation("="));
		expression.setType(getType("Boolean"));
	
		// Create type for attribute (and for the attribute call expression).
		MClassifier attributeType = new MClassifierImpl();
		attributeType.setName(attribute.getType());
		
		// First create the left hand side of the operation. This navigates to 
		// the attribute.
		MAttributeCallExp attrCallExp = new MAttributeCallExpImpl();
		attrCallExp.setParentOperation(expression);
		attrCallExp.setSource(convertNavigationPath(path));
		attrCallExp.setType(attributeType);
		MAttribute attrib = new MAttributeImpl();
		attrib.setName(attribute.getName());
		attrib.setType(attributeType);
		attrCallExp.setReferredAttribute(attrib);
		expression.addArguments(attrCallExp);
		
		// Create right hand side.
		VariableDeclaration varDecl = attribute.getVariableDeclaration();
		if (varDecl == null) {
			MLiteralExp valueExp = null;
			if (attribute.getType().equals("Boolean"))
				valueExp = new MBooleanLiteralExpImpl();
			else if (attribute.getType().equals("Integer"))
				valueExp = new MIntegerLiteralExpImpl();
			else if (attribute.getType().equals("Real"))
				valueExp = new MRealLiteralExpImpl();
			else if (attribute.getType().equals("String"))
				valueExp = new MStringLiteralExpImpl();
			else
				valueExp = new MLiteralExpImpl();
			valueExp.setParentOperation(expression);
			valueExp.setType(attributeType);
			valueExp.setName(attribute.getValue());
			expression.addArguments(valueExp);
		} else {
			MVariableDeclaration oclVarDecl =
				getVariableDeclaration(varDecl.getVarName());
			oclVarDecl.setType(getType(varDecl.getType()));
			MVariableExp varExp = new MVariableExpImpl();
			varExp.setType(oclVarDecl.getType());
			varExp.setReferredVariable(oclVarDecl);
			expression.addArguments(varExp);
		}
		
		return expression;	
	}
	
	/**
	 * 
	 */
	protected MOclExpression convertMethodCall(NavigationPath path) {
		Method method = ((ClassifierRole)path.getLastNode()).getMethod();
		
		// Root operation
		MOperationCallExp opCallExp = new MOperationCallExpImpl();
		opCallExp.setSource(convertNavigationPath(path));
		
		// Create operation.
		MOperation operation = new MOperationImpl();
		operation.setName(method.getName());
		opCallExp.setReferredOperation(operation);
		
		// Create parameters.
		for (Iterator paramIt = method.getCall().iterator(); 
			 paramIt.hasNext(); ) {
			VariableDeclaration varDecl = 
				(VariableDeclaration)paramIt.next();
				
			MAttributeCallExp attrCallExp = new MAttributeCallExpImpl();
				
			MClassifier attrType = new MClassifierImpl();
			attrType.setName(varDecl.getType());
			attrCallExp.setType(attrType);
				
			MAttribute attr = new MAttributeImpl();
			attr.setName(varDecl.getVarName());
			attr.setType(attrType);
				
			attrCallExp.setReferredAttribute(attr);
			opCallExp.addArguments(attrCallExp); 
		}
		
		// Return variable or not?
		if (method.getReturns() == null) {
			operation.setQuery(true);
			opCallExp.setType(getType("Boolean"));
		} else {
			operation.setQuery(false);
			VariableDeclaration varDecl = method.getReturns();
			VariableDeclaration roleVar = method.getClassifierRole().getClassifierVar();
			
			MOperationCallExp expression = new MOperationCallExpImpl();
			expression.setReferredOperation(getOperation("="));
			expression.setType(getType("Boolean"));
			
			if (varDecl.getVarName().equals("result") &&
			    roleVar != null && 
			    roleVar.getVarName().equals("self")) {
				MVariableDeclaration oclVarDecl =
					getVariableDeclaration(varDecl.getVarName());
				oclVarDecl.setType(getType(varDecl.getType()));
				MVariableExp varExp = new MVariableExpImpl();
				varExp.setType(oclVarDecl.getType());
				varExp.setReferredVariable(oclVarDecl);
				expression.addArguments(varExp);
			} else	
				expression.addArguments(opCallExp);
			
			MVariableDeclaration oclVarDecl =
				getVariableDeclaration(varDecl.getVarName());
			oclVarDecl.setType(getType(varDecl.getType()));
			MVariableExp varExp = new MVariableExpImpl();
			varExp.setType(oclVarDecl.getType());
			varExp.setReferredVariable(oclVarDecl);
			expression.addArguments(varExp);
			
			opCallExp.setType(varExp.getType());
			
			opCallExp = expression;
		}
		
		return opCallExp;	
	}

	/**
	 * 
	 */
	protected MOclExpression convertEmptyOperation(NavigationPath path) {
		if (!path.isEmptyOperation())
			return null;
	
		MOperationCallExp opCallExp = new MOperationCallExpImpl();
		
		MAssociationEndCallExp nav = 
			(MAssociationEndCallExp)convertNavigationPath(path);
		MAssociationEnd navEnd = nav.getReferredAssociationEnd();
			
		MSetType endSetType = new MSetTypeImpl();
		endSetType.setElementType(navEnd.getType());
		nav.setType(endSetType);
		
		opCallExp.setSource(nav);
			
		// Create operation.
		MOperation operation = new MOperationImpl();
		operation.setName("empty");
		opCallExp.setReferredOperation(operation);
		opCallExp.setType(getType("Boolean"));
			
		return opCallExp;
	}

	/**
	 * 
	 */
	protected MOclExpression convertNotEmptyOperation(NavigationPath path) {
		if (!path.isNotEmptyOperation())
			return null;
	
		MOperationCallExp opCallExp = new MOperationCallExpImpl();
		
		MAssociationEndCallExp nav = 
			(MAssociationEndCallExp)convertNavigationPath(path);
		MAssociationEnd navEnd = nav.getReferredAssociationEnd();
			
		MSetType endSetType = new MSetTypeImpl();
		endSetType.setElementType(navEnd.getType());
		nav.setType(endSetType);
		
		opCallExp.setSource(nav);
			
		// Create operation.
		MOperation operation = new MOperationImpl();
		operation.setName("notEmpty");
		opCallExp.setReferredOperation(operation);
		opCallExp.setType(getType("Boolean"));
			
		return opCallExp;
	}
	
	/**
	 * 
	 */
	protected MOclExpression convertSizeOperation(NavigationPath path) {
		if (!path.isSizeOperation())
			return null;
		
		// Root operation.
		MOperationCallExp oclExp = new MOperationCallExpImpl();
		oclExp.setType(getType("Boolean"));
		oclExp.setReferredOperation(getOperation("="));
		
		// Left hand side.
		MOperationCallExp leftExp = new MOperationCallExpImpl();
		
		MAssociationEndCallExp nav = 
			(MAssociationEndCallExp)convertNavigationPath(path);
		MAssociationEnd navEnd = nav.getReferredAssociationEnd();
			
		MSetType endSetType = new MSetTypeImpl();
		endSetType.setElementType(navEnd.getType());
		nav.setType(endSetType);
		
		leftExp.setSource(nav);
			
		// Create operation.
		MOperation operation = new MOperationImpl();
		operation.setName("size");
		leftExp.setReferredOperation(operation);
		leftExp.setType(getType("Integer"));
		
		// Create right side.
		MVariableExp rightExp = new MVariableExpImpl();
		rightExp.setType(getType("Integer"));
		ClassifierRoleSet roleSet = (ClassifierRoleSet)path.getLastNode();
		MVariableDeclaration rightVar = 
			getVariableDeclaration(roleSet.getOpKindEnum().getCallVar().getVarName());
		rightVar.setType(getType("Integer"));
		rightExp.setReferredVariable(rightVar);
			
		// Add left and right hand side.
		oclExp.addArguments(leftExp);
		oclExp.addArguments(rightExp);
			
		return oclExp;
	}
	
	/**
	 * 
	 */
	protected MOclExpression convertSumOperation(NavigationPath path) {
		if (!path.isSumOperation())
			return null;
		
		ClassifierRoleSet roleSet = (ClassifierRoleSet)path.getLastNode();
		
		// Get attribute which matches the iterator variable.
		Attribute attrib = null;
		String iteratorVarName = 
			roleSet.getOpKindEnum().getIteratorVar().getVarName();
		for (Iterator it = path.getLastNode().getAttribute().iterator();
		     it.hasNext(); ) {
			Attribute curAttrib = (Attribute)it.next();
			String attrVarName = curAttrib.getVariableDeclaration().getVarName();
			if (iteratorVarName.equals(attrVarName)) {
				attrib = curAttrib;
				break;
			}
		}
		
		if (attrib == null)
			return null;
		
		// Root operation.
		MOperationCallExp oclExp = new MOperationCallExpImpl();
		oclExp.setType(getType("Boolean"));
		oclExp.setReferredOperation(getOperation("="));
		
		// Left hand side.
		MOperationCallExp leftExp = new MOperationCallExpImpl();
		leftExp.setType(getType("Integer"));
		leftExp.setReferredOperation(getOperation("sum"));	
		
		MAttributeCallExp attrCall = new MAttributeCallExpImpl();
		leftExp.setSource(attrCall);
		MAttribute oclAttrib = new MAttributeImpl();
		oclAttrib.setName(attrib.getName());
		oclAttrib.setType(getType(attrib.getType()));
		attrCall.setReferredAttribute(oclAttrib);
		
		MSetType endSetType = new MSetTypeImpl();
		endSetType.setElementType(getType(attrib.getType()));
		attrCall.setType(endSetType);
		
		attrCall.setSource(convertNavigationPath(path));	
			
		// Create operation.
		MOperation operation = new MOperationImpl();
		operation.setName("sum");
		leftExp.setReferredOperation(operation);
		leftExp.setType(getType("Integer"));
		
		// Create right side.
		MVariableExp rightExp = new MVariableExpImpl();
		rightExp.setType(getType("Integer"));
		MVariableDeclaration rightVar = 
			getVariableDeclaration(roleSet.getOpKindEnum().getCallVar().getVarName());
		rightVar.setType(getType("Integer"));
		rightExp.setReferredVariable(rightVar);
			
		// Add left and right hand side.
		oclExp.addArguments(leftExp);
		oclExp.addArguments(rightExp);
			
		return oclExp;
	}
	
	/**
	 * Returns a list of association roles where the given classifier node is
	 * a source node.
	 * 
	 * @param node
	 *     The classifier node to return the list for.
	 * @return
	 *     A list of AssociationRole instances representing the described 
	 *     association roles.
	 */
	protected static List getOutgoingAssociations(ClassifierNode node) {
		Vector associations = new Vector();
		for (Iterator it = node.getAssociationEndRole().iterator(); 
		     it.hasNext(); ) {
			AssociationEndRole role = (AssociationEndRole)it.next();
			if (role instanceof AssociationEndRoleSource)
				associations.add(((AssociationEndRoleSource)role).getAssociationRole());
		}
		return associations;
	}

	/**
	 * Concatenates two given MOclExpressions via "and". 
	 *
	 * @param expression1
	 *     The first expression. May be <tt>null</tt>. In this case,  
	 *     expression2 as single expression will be returned.
	 * @param expression2
	 *     The second expression. May be <tt>null</tt>. In this case,
	 *     expression1 will be returned.
	 * 
	 * @return
	 *     The concatenation.
	 */
	protected MOclExpression concatenate(MOclExpression expression1,
										 MOclExpression expression2) {
		if (expression1 == null)
			return expression2;
		else if (expression2 == null)
			return expression1;
		else {
			MOperationCallExp exp = new MOperationCallExpImpl();
			exp.setType(getType("Boolean"));
			exp.setReferredOperation(getOperation("and"));
			exp.addArguments(expression1);
			exp.addArguments(expression2);
			return exp;
		}
	}

	/**
	 * Searches in the given constraint a classifier node with the given name.
	 * 
	 * @param constraint
	 *     The constraint to search the classifier node in.
	 * @param name
	 *     The name of the classifier node to search for.
	 * @return
	 *     The classifier node with the given name or <tt>null</tt>, if this was
	 *     not found.
	 */
	protected static ClassifierNode findClassifierNode(NavExp navExp, String name) {
		List nodes = navExp.getClassifierNode();
		for (Iterator it = nodes.iterator(); it.hasNext(); ) {
			ClassifierNode node = (ClassifierNode)it.next();
			if (name.equals(node.getName()))
				return node;
		}				
		
		return null;
	}

	// Returns null, if not found
//	protected Method findMethod(Constraint constraint, String classifierName, String methodName) {
//		Stack stack = new Stack();
//		if (constraint.getVOCLExpression() != null)
//			stack.add(constraint.getVOCLExpression());
//		while (!stack.isEmpty()) {
//			VOCLExpression exp = (VOCLExpression)stack.pop();
//			if (exp instanceof NavExp) {
//				List roles = ((NavExp)exp).getClassifierNode();
//				for (Iterator it = roles.iterator(); it.hasNext(); ) {
//					Object obj = it.next();
//					if (obj instanceof ClassifierRole) {
//						ClassifierRole role = (ClassifierRole)obj;
//						if (classifierName.equals(role.getType())) {
//							Method method = role.getMethod();
//							if (method != null && methodName.equals(method.getName()))
//								return method;
//						}
//					}
//				}
//			} else if (exp instanceof BoolExp) {
//				stack.add(((BoolExp)exp).getFirstOp());
//				stack.add(((BoolExp)exp).getSecondOp());
//			} else if (exp instanceof IfExp) {
//				stack.add(((IfExp)exp).getIfCondition());
//				stack.add(((IfExp)exp).getThenExp());
//				stack.add(((IfExp)exp).getElseExp());
//			} else if (exp instanceof LetExp) {
//				stack.add(((LetExp)exp).getDef());
//				stack.add(((LetExp)exp).getInExp());
//			}
//		}
//		
//		return null;
//	}

	protected static List getVariablesFromMethod(String method) {
		List vars = new Vector();
		
		int paramBegin = method.indexOf('(');
		int paramEnd = method.indexOf(')');
		if (paramBegin != -1 && paramEnd != -1) {
			String params = method.substring(paramBegin + 1, paramEnd);
			StringTokenizer tok = new StringTokenizer(params, ",");
			while (tok.hasMoreTokens()) {
				String token = tok.nextToken().trim();
				vars.add(token);
			}
		}
		
		return vars;
	}
	
	protected static String getVarName(String var) {
		int splitPos = var.indexOf(':');
		return var.substring(0, splitPos);
	}
	
	protected static String getVarType(String var) {
		int splitPos = var.indexOf(':');
		return var.substring(splitPos + 1);
	}
}
