/* 
 * $Id: ModelValidator.java,v 1.2 2004/11/09 09:49:37 haensgen Exp $
 */
 
package VoclToOclConverter;

import java.util.Iterator;
import java.util.List;

import VOCLModel.*;

/**
 * Validates a given EMF model instance, if it represents a correct Visual OCL 
 * expression. For example, it will be checked whether all classifier roles are
 * typed, whether all association roles have an unnamed and a named end, if 
 * those ends are typed etc.
 */
public class ModelValidator {

	// Root of the EMF model instance.
	private VOCLDiagram modelRoot = null;
	
	// To collect all invalid elements.
	private BadElementCollector invalidElements = null;

	/**
	 * Creates a new ModelValidator instance for the EMF model instance whose
	 * root is given.
	 * 
	 * @param modelRoot
	 *     The root of the EMF model instance.
	 */
	public ModelValidator(VOCLDiagram modelRoot) {
		this.modelRoot = modelRoot;
		invalidElements = new BadElementCollector(modelRoot, 100);
	}
	
	/**
	 * Returns the list of states of invalid element determined during 
	 * isValid(). This list contains {@link ElementStatus} instances.
	 * 
	 * @return
	 *     A list of ElementStatus instances which represent states of invalid
	 *     elements.
	 */
	public List getInvalidElementStates() {
		return invalidElements.getElements();
	}
	
	/**
	 * Checks whether the model is valid.
	 * 
	 * @return
	 *     <tt>true</tt> if it is valid. <tt>false</tt> if not. In the last
	 *     case call <tt>getInvalidElementStates()</tt> to get the list of 
	 *     states of invalid elements.
	 * 
	 * @see #getInvalidElementStates()
	 */
	public boolean isValid() {
		boolean valid = true;
		
		invalidElements.reset();
		
		// Check all constraints
		List constraints = modelRoot.getConstraint();
		if (constraints != null && constraints.size() > 0) {
			for (Iterator it = constraints.iterator(); it.hasNext(); )
				if (!checkConstraint((Constraint)it.next()))
					valid = false;
		}
		
		return valid;
	}
	
	/**
	 * Checks a constraint whether it is valid. A constraint is valid, if
	 * <ul>
	 *     <li>there is a valid context,</tt>
	 *     <li>all assigned conditions are valid,</li>
	 *     <li>all variables are valid, and</li>
	 *     <li>there is one unique VOCL expression.</li>
	 * </ul>
	 * 
	 * @param constraint
	 *     The constraint to check.
	 * @return 
	 *     <tt>true</tt> if it is valid. <tt>false</tt> if not. 
	 */
	protected boolean checkConstraint(Constraint constraint) {
		boolean valid = true;

		// Check for context
		Context context = constraint.getConstraintContext();
		if (context != null) {
			if (!checkContext(context))			
				valid = false;
		} else {
			invalidElements.addElement(
				IElementStatusConstants.CONSTRAINT__NO_CONTEXT,
				"Constraint \"" + constraint.getName() + "\" has no " +
					"context assigned.",
				constraint);
			valid = false;			
		}
		
		// Check condition if there is any
		Condition condition = constraint.getCondition();
		if (condition != null)
			if (!checkCondition(condition))			
				valid = false;
				
		// Check all variables
		/*
		List varDecls = constraint.getVariableDeclaration();
		if (varDecls != null) {
			for (Iterator varIt = varDecls.iterator(); varIt.hasNext(); ) {
				VariableDeclaration varDecl = (VariableDeclaration)varIt.next();
				if (!checkVariableDeclaration(varDecl))
					valid = false;
			}
		}
		*/
				
		// Check if there is one unique VOCL expression
		VOCLExpression exp = constraint.getVOCLExpression();
		if (exp != null) {
			// TODO Check VOCL expression
			//if (!checkNavExp(navExp))			
			//	valid = false;	
		} 
		
		return valid;
	}
	
	/**
	 * Checks a context, if it is valid. A context is valid, if
	 * <ul>
	 *     <li>it is typed, i.e. the type string is not empty.</li>
	 * </ul>
	 * 
	 * @param context
	 *     The context to check.
	 * @return 
	 *     <tt>true</tt> if it is valid. <tt>false</tt> if not. 
	 */
	protected boolean checkContext(Context context) {
		boolean valid = true;
		
		// Check type string
		String type = context.getType();
		if (type == null || type.trim().length() == 0) {
			invalidElements.addElement(
				IElementStatusConstants.CONTEXT__NO_TYPE,
				"Context has no type.",
				context);
			valid = false;				
		}
		
		return valid;
	}
	
	/**
	 * Checks a condition, if it is valid. A condition is valid, if
	 * <ul>
	 *     <li>its operation name is not empty, and</li>
	 *     <li>it has a "left" variable declaration.</li>
	 * </ul>
	 * 
	 * @param condition
	 *     The condition to check.
	 * @return 
	 *     <tt>true</tt> if it is valid. <tt>false</tt> if not. 
	 */
	protected boolean checkCondition(Condition condition) {
		boolean valid = true;
		
		// Check operation name. This must not be empty.
		String op = condition.getOp();
		if (op == null || op.trim().equals("")) {
			invalidElements.addElement(
				IElementStatusConstants.CONDITION__NO_OPERATION,
				"Condition of constraint \"" + 
					condition.getConstraint().getName() + "\" has no " +
					"operation assigned.",
				condition);
			valid = false;
		}
	
		return valid;
	}
	
	/**
	 * Checks a variable declaration, if it is valid. Such a declaration is 
	 * valid, if
	 * <ul>
	 *     <li>it is typed, and</li>
	 *     <li>has a non-empty name.</li>
	 * </ul>
	 * 
	 * @param variableDeclaration
	 *     The variable declaration to check.
	 * @return 
	 *     <tt>true</tt> if it is valid. <tt>false</tt> if not. 
	 */
	protected boolean checkVariableDeclaration(VariableDeclaration variableDeclaration) {
		boolean valid = true;
	
		// Check type.
		String type = variableDeclaration.getType();
		if (type == null || type.trim().length() == 0) {
			invalidElements.addElement(
				IElementStatusConstants.VARIABLE_DECLARATION__NO_TYPE,
				"Found variable declaration with no type.",
				variableDeclaration);
			valid = false;				
		}
		
		// Check name.
		String name = variableDeclaration.getVarName();
		if (name == null || name.trim().length() == 0) {
			invalidElements.addElement(
				IElementStatusConstants.VARIABLE_DECLARATION__NO_NAME,
				"Found variable declaration of type \"" + type + "\" with " +
					"no name.",
				variableDeclaration);
			valid = false;				
		}
	
		return valid;
	}
	
	/**
	 * Checks a navigation expression, if it is valid. Such an expression is 
	 * valid, if
	 * <ul>
	 *     <li>all variable declarations are valid,</li>
	 *     <li>there is either a "self" variable or a variable named like the
	 *         variable associated with the context of the constraint,</li>
	 *     <li>all classifier roles are valid, and</li>
	 *     <li>all association roles are valid.</li>
	 * </ul>
	 * 
	 * @param navExp
	 *     The navigation expression to check.
	 * @return 
	 *     <tt>true</tt> if it is valid. <tt>false</tt> if not. 
	 */
	protected boolean checkNavExp(NavExp navExp) {
		boolean valid = true;
	
		// Check variable declarations. Variables must have a non-empty name 
		// and a type.
		/*
		List varDecls = navExp.getVariableDeclaration();
		if (varDecls != null) {
			for (Iterator it = varDecls.iterator(); it.hasNext(); ) 
				if (!checkVariableDeclaration((VariableDeclaration)it.next()))
					valid = false;
		}
		*/
		
		// Check for "self".
		/*
		String name = "self";
		varDecls = navExp.getConstraint().getConstraintContext().getVariableDeclaration();
		if (varDecls != null && varDecls.size() > 0)
			name = ((VariableDeclaration)varDecls.get(0)).getVarName();
		varDecls = navExp.getVariableDeclaration();
		if (varDecls != null) {
			int selfCounter = 0;
			for (Iterator it = varDecls.iterator(); it.hasNext(); ) {
				VariableDeclaration varDecl = (VariableDeclaration)it.next();
				if (name.equals(varDecl.getVarName()))
					selfCounter++;
			}
			if (selfCounter == 0) {
				addInvalidElement(
					IElementStatusConstants.NAVEXP__NO_SELF_VARIABLE_DECLARATION,
					"Navigation expression has no self variable declaration.",
					navExp);
				valid = false;	
			}
			else if (selfCounter > 1) {
				addInvalidElement(
					IElementStatusConstants.NAVEXP__MORE_THAN_ONE_SELF_VARIABLE_DECLARATION,
					"Navigation expression has more than one self variable declaration.",
					navExp);
				valid = false;	
			}
		} else {
			addInvalidElement(
				IElementStatusConstants.NAVEXP__NO_SELF_VARIABLE_DECLARATION,
				"Navigation expression has no self variable declaration.",
				navExp);
			valid = false;			
		}
		*/
		
		// Check classifier roles.
		/*
		List classifierRoles = navExp.getClassifierRoleNode();
		if (classifierRoles != null) {
			for (Iterator it = classifierRoles.iterator(); it.hasNext(); ) 
				if (!checkClassifierRole((ClassifierRole)it.next()))
					valid = false;
		}
		*/
		
		// Check association roles.
		List associationRoles = navExp.getAssociationRole();
		if (associationRoles != null) {
			for (Iterator it = associationRoles.iterator(); it.hasNext(); ) 
				if (!checkAssociationRole((AssociationRole)it.next()))
					valid = false;
		}
	
		return valid;
	}
	
	/**
	 * Checks a classifier role, if it is valid. A classifier role is valid, if 
	 * <ul>
	 * </ul>
	 * 
	 * @param classifierRole
	 *     The classifier role to check.
	 * @return 
	 *     <tt>true</tt> if it is valid. <tt>false</tt> if not. 
	 */
	protected boolean checkClassifierRole(ClassifierRole classifierRole) {
		boolean valid = true;
	
		return valid;
	}
	
	/**
	 * Checks a association role, if it is valid. An association role is valid, 
	 * if
	 * <ul>
	 * </ul>
	 * 
	 * @param associationRole
	 *     The association role to check.
	 * @return 
	 *     <tt>true</tt> if it is valid. <tt>false</tt> if not. 
	 */
	protected boolean checkAssociationRole(AssociationRole associationRole) {
		boolean valid = true;
	
		return valid;
	}
}
