/*
 * $Id: NavigationPath.java,v 1.2 2004/11/09 09:49:37 haensgen Exp $
 */
 
package VoclToOclConverter;

import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import VOCLModel.*;

/**
 * Represents a navigation path in a Visual OCL expression.
 */
public class NavigationPath {

	// The path. It contains, in alternating order, classifier nodes and 
	// association roles. The first and the last element are classifier nodes.
	private Vector path;

	/**
	 * Creates a new empty path.
	 * 
	 * @param firstNode
	 *     The classifier node from which to start the path.
	 */
	public NavigationPath(ClassifierNode firstNode) {
		path = new Vector();
		path.add(firstNode);
	}
	
	/**
	 * Creates a copy of the given path.
	 * 
	 * @param pathToCopy
	 *     The path to create a copy from.
	 */
	public NavigationPath(NavigationPath pathToCopy) {
		path = new Vector();
		for (Iterator it = pathToCopy.iterator(); it.hasNext(); ) 
			path.add(it.next());
	}
	
	/**
	 * Returns an iterator to iterate through the elements of the path. A path 
	 * contains, in alternating order, of classifier nodes (instances of 
	 * VOCLModel.ClassifierNode) and association roles (instances of 
	 * VOCLModel.AssociationRole). The first and the last elements are 
	 * classifier nodes.
	 * 
	 * @return
	 *     The navigation path iterator.
	 */
	public Iterator iterator() {
		return path.iterator();
	}
	
	/**
	 * Returns the length of the path, i.e. the number of association roles
	 * used on the path.
	 * 
	 * @return  
	 *     The length of the path.
	 */
	public int getLength() {
		return (path.size() - 1) / 2;
	}

	/**
	 * Returns the classifier node from which the path starts.
	 * 
	 * @return 
	 *     The first classifier node.
	 */
	public ClassifierNode getFirstNode() {
		return (ClassifierNode)path.firstElement();
	}
	
	/**
	 * Returns the classifier node where the path ends.
	 *
	 * @return 
	 *     The last classifier node.
	 */
	public ClassifierNode getLastNode() {
		return (ClassifierNode)path.lastElement();
	}
	
	/**
	 * Extends the path by a new association role and the corresponding target 
	 * classifier node. 
	 * 
	 * @param associationRole
	 *     The association role. It is assumed that the type of the source 
	 *     association end classifier node is equal to the last classifier node 
	 *     of the current path.
	 * 
	 * @exception IncompatibleAssociationRoleException
	 *     If the type of the source association end classifier node is not 
	 *     equal to the last classifier node of the current path.
	 *     
	 */
	public void add(AssociationRole associationRole) throws 
		IncompatibleAssociationRoleException {
		ClassifierNode sourceType = 
			associationRole.getAssociationEndRoleSource().getClassifierNode();
		ClassifierNode targetType =
			associationRole.getAssociationEndRoleTarget().getClassifierNode();
		if (!sourceType.equals(getLastNode()))
			throw new IncompatibleAssociationRoleException(sourceType  +
				" is incompatible with " + getLastNode());
		path.add(associationRole);
		path.add(targetType);
	}
	
	/**
	 * Tests whether the given path is equivalent to the current one, i.e. that 
	 * they start and end at the same classifier nodes and use different
	 * association roles.
	 * 
	 * @param path
	 *     The path to compare to this one.
	 * @return
	 *     <tt>true</tt>, if the two paths are equivalent. Otherwise 
	 *     <tt>false</tt>.
	 */
	public boolean isEquivalent(NavigationPath path) {
		if (!isAcyclic())
			return false;
		if (!getFirstNode().equals(path.getFirstNode()) || 
		    !getLastNode().equals(path.getLastNode()))
			return false;
		else if (getLength() != path.getLength())
			return true;
		else {
			for (Iterator it = path.iterator(); it.hasNext(); ) {
				Object obj = it.next();
				if (obj instanceof AssociationRole && 
			    	!this.path.contains((AssociationRole)obj))
					return true;
			}
		}
		return false;
	}
	
	/**
	 * Checks whether the path contains the given classifier node.
	 * 
	 * @param classifierNode
	 *     The classifier node.
	 * @return
	 *     <tt>true</tt>, if it contains the given classifier node. Otherwise 
	 *     <tt>false</tt>.
	 */
	public boolean contains(ClassifierNode classifierNode) {
		return path.contains(classifierNode);
	}
	
	/**
	 * Checks whether the path contains the given association role.
	 * 
	 * @param associationRole
	 *     The association role.
	 * @return
	 *     <tt>true</tt>, if it contains the given association role. Otherwise 
	 *     <tt>false</tt>.
	 */
	public boolean contains(AssociationRole associationRole) {
		return path.contains(associationRole);
	}

	/**
	 * Checks whether the path is a round trip, i.e. a path which ends in the 
	 * first classifier node.
	 * 
	 * @return 
	 *     <tt>true</tt>, if it is a round trip. Otherwise <tt>false</tt>.
	 */
	public boolean isRoundTrip() {
		return getLength() > 0 && getFirstNode().equals(getLastNode());
	}
	
	/**
	 * Checks whether the path is acyclic, i.e. whether the last node has no 
	 * outgoing edges.
	 * 
	 * @return
	 *     <tt>true</tt>, if it is acyclic. Otherwise <tt>false</tt>.
	 */
	public boolean isAcyclic() {
		ClassifierNode node = getLastNode();
		
		// Check whether node has outgoing edges.
		boolean hasOutgoingEdges = false;
		for (Iterator it = node.getAssociationEndRole().iterator(); 
			 !hasOutgoingEdges && it.hasNext(); ) {
			AssociationEndRole endRole = (AssociationEndRole)it.next();
			if (endRole instanceof AssociationEndRoleSource &&
				endRole.getClassifierNode().equals(node)) 
				hasOutgoingEdges = true;
		}
		
		return !hasOutgoingEdges;
	}
	
	/**
	 * Checks whether the path is an attribute navigation, i.e. whether it 
	 * reaches a classifier role with attributes assigned to it.
	 *
	 * @return 
	 *     <tt>true</tt>, if the path reaches such a classifier role. Otherwise
	 *     <tt>false</tt>.
	 */
	public boolean isAttributeCall() {
		boolean attributeCall = false;
		
		if (getLastNode() instanceof ClassifierRole) {
			List attributes = getLastNode().getAttribute();
			attributeCall = (attributes != null) && (attributes.size() > 0);
		}
		
		return attributeCall;
	}
	
	/**
	 * TODO: Kommentar hinzufgen
	 */
	public boolean isMethodCall() {
		ClassifierNode lastNode = getLastNode();
		if (lastNode instanceof ClassifierRole)
			return ((ClassifierRole)lastNode).getMethod() != null;
		else
			return false;
	}
	
	/**
	 * Checks whether the path represents the operation "EMPTY" on a set.
	 * 
	 * @return 
	 *     <tt>true</tt>, if the path represents this operation. Otherwise
	 *     <tt>false</tt>.
	 */
	public boolean isEmptyOperation() {
		boolean emptyOperation = false;
		
		if (getLastNode() instanceof ClassifierRoleSet) {
			ClassifierRoleSet set = (ClassifierRoleSet)getLastNode();
			String kind = set.getOperationKind().trim().toUpperCase();
			emptyOperation = kind.equals("EMPTY");
		}
		
		return emptyOperation;
	}
	
	/**
	 * Checks whether the path represents the operation "NOT_EMPTY" on a set.
	 * 
	 * @return 
	 *     <tt>true</tt>, if the path represents this operation. Otherwise
	 *     <tt>false</tt>.
	 */
	public boolean isNotEmptyOperation() {
		boolean notEmptyOperation = false;
		
		if (getLastNode() instanceof ClassifierRoleSet) {
			ClassifierRoleSet set = (ClassifierRoleSet)getLastNode();
			String kind = set.getOperationKind().trim().toUpperCase();
			notEmptyOperation = kind.equals("NOT_EMPTY");
		}
		
		return notEmptyOperation;
	}
	
	/**
	 * Checks whether the path represents the operation "SIZE" on a set.
	 * 
	 * @return 
	 *     <tt>true</tt>, if the path represents this operation. Otherwise
	 *     <tt>false</tt>.
	 */
	public boolean isSizeOperation() {
		boolean sizeOperation = false;
		
		if (getLastNode() instanceof ClassifierRoleSet) {
			ClassifierRoleSet set = (ClassifierRoleSet)getLastNode();
			String kind = set.getOperationKind().trim().toUpperCase();
			sizeOperation = kind.equals("SIZE");
		}
		
		return sizeOperation;
	}
	
	/**
	 * Checks whether the path represents the operation "SUM" on a set.
	 * 
	 * @return 
	 *     <tt>true</tt>, if the path represents this operation. Otherwise
	 *     <tt>false</tt>.
	 */
	public boolean isSumOperation() {
		boolean sumOperation = false;
		
		if (getLastNode() instanceof ClassifierRoleSet) {
			ClassifierRoleSet set = (ClassifierRoleSet)getLastNode();
			String kind = set.getOperationKind().trim().toUpperCase();
			sumOperation = kind.equals("SUM");
		}
		
		return sumOperation;
	}
	
	/**
	 * Returns a string representation of the path.
	 * 
	 * @return
	 *     A string which represents the path.
	 */
	public String toString() {
		String s = "";
		
		VariableDeclaration varDecl = getFirstNode().getClassifierVar();
		if (varDecl != null)
			s += varDecl.getVarName();
		
		// Add the sequence of association end roles used on the path 
		Iterator it = path.iterator();
		it.next(); // ignore first classifier role
		while (it.hasNext()) {
			AssociationRole role = (AssociationRole)it.next();
			s += "." + role.getAssociationEndRoleTarget().getName();
			it.next(); // ignore following classifier role
		}
		
		return "[" + s + "]";
	}
}
